#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from botocore.exceptions import ClientError, WaiterError
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.ec2 import EC2Hook
from airflow.providers.amazon.aws.operators.ec2 import (
    EC2CreateInstanceOperator,
    EC2HibernateInstanceOperator,
    EC2RebootInstanceOperator,
    EC2StartInstanceOperator,
    EC2StopInstanceOperator,
    EC2TerminateInstanceOperator,
)
from airflow.providers.common.compat.sdk import AirflowException

from unit.amazon.aws.utils.test_template_fields import validate_template_fields


class BaseEc2TestClass:
    @classmethod
    def _get_image_id(cls, hook):
        """Get a valid image id to create an instance."""
        conn = hook.get_conn()
        try:
            ec2_client = conn.meta.client
        except AttributeError:
            ec2_client = conn

        # We need an existing AMI Image ID otherwise `moto` will raise DeprecationWarning.
        images = ec2_client.describe_images()["Images"]
        return images[0]["ImageId"]


class TestEC2CreateInstanceOperator(BaseEc2TestClass):
    def test_init(self):
        ec2_operator = EC2CreateInstanceOperator(
            task_id="test_create_instance",
            image_id="test_image_id",
        )

        assert ec2_operator.task_id == "test_create_instance"
        assert ec2_operator.image_id == "test_image_id"
        assert ec2_operator.max_count == 1
        assert ec2_operator.min_count == 1
        assert ec2_operator.max_attempts == 20
        assert ec2_operator.poll_interval == 20

    @mock_aws
    def test_create_instance(self):
        ec2_hook = EC2Hook()
        create_instance = EC2CreateInstanceOperator(
            image_id=self._get_image_id(ec2_hook),
            task_id="test_create_instance",
        )
        instance_id = create_instance.execute(None)

        assert ec2_hook.get_instance_state(instance_id=instance_id[0]) == "running"

    @mock_aws
    def test_create_multiple_instances(self):
        ec2_hook = EC2Hook()
        create_instances = EC2CreateInstanceOperator(
            task_id="test_create_multiple_instances",
            image_id=self._get_image_id(hook=ec2_hook),
            min_count=5,
            max_count=5,
        )
        instance_ids = create_instances.execute(None)
        assert len(instance_ids) == 5

        for id in instance_ids:
            assert ec2_hook.get_instance_state(instance_id=id) == "running"

    def test_template_fields(self):
        ec2_operator = EC2CreateInstanceOperator(
            task_id="test_create_instance",
            image_id="test_image_id",
        )
        validate_template_fields(ec2_operator)

    @mock_aws
    def test_cleanup_on_post_create_failure(self):
        ec2_hook = EC2Hook()

        operator = EC2CreateInstanceOperator(
            task_id="test_cleanup_on_error",
            image_id=self._get_image_id(ec2_hook),
            wait_for_completion=True,
        )

        waiter_error = WaiterError(
            "InstanceRunning",
            "You are not authorized to perform this operation",
            {},
        )

        # Force failure after instance creation (e.g. missing DescribeInstances permission).
        with mock.patch.object(operator.hook, "get_waiter") as mock_get_waiter:
            mock_get_waiter.return_value.wait.side_effect = waiter_error
            with pytest.raises(WaiterError) as exc:
                operator.execute(None)

        # Ensure the original waiter exception is propagated unchanged.
        assert exc.value is waiter_error

        # Instance must have been terminated.
        # We know exactly one instance was created.
        instances = list(ec2_hook.conn.instances.all())
        assert len(instances) == 1

        instance = instances[0]
        assert instance.state["Name"] == "terminated"

    @mock_aws
    def test_cleanup_failure_propagates_original_exception(self):
        ec2_hook = EC2Hook()

        operator = EC2CreateInstanceOperator(
            task_id="test_cleanup_failure_does_not_mask_error",
            image_id=self._get_image_id(ec2_hook),
            wait_for_completion=True,
        )

        waiter_error = WaiterError(
            "InstanceRunning",
            "You are not authorized to perform this operation",
            {},
        )

        cleanup_error = ClientError(
            error_response={
                "Error": {
                    "Code": "UnauthorizedOperation",
                    "Message": "You are not authorized to perform this operation",
                }
            },
            operation_name="TerminateInstances",
        )

        with (
            mock.patch.object(operator.hook, "get_waiter") as mock_get_waiter,
            mock.patch.object(operator.hook, "terminate_instances") as mock_terminate,
        ):
            mock_get_waiter.return_value.wait.side_effect = waiter_error
            mock_terminate.side_effect = cleanup_error

            with pytest.raises(WaiterError) as exc:
                operator.execute(None)

            # Ensure the original waiter exception is propagated unchanged.
            assert exc.value is waiter_error

        # Cleanup is best-effort; failure to terminate must not override the original error.


class TestEC2TerminateInstanceOperator(BaseEc2TestClass):
    def test_init(self):
        ec2_operator = EC2TerminateInstanceOperator(
            task_id="test_terminate_instance",
            instance_ids="test_image_id",
        )

        assert ec2_operator.task_id == "test_terminate_instance"
        assert ec2_operator.max_attempts == 20
        assert ec2_operator.poll_interval == 20

    @mock_aws
    def test_terminate_instance(self):
        ec2_hook = EC2Hook()

        create_instance = EC2CreateInstanceOperator(
            image_id=self._get_image_id(ec2_hook),
            task_id="test_create_instance",
        )
        instance_id = create_instance.execute(None)

        assert ec2_hook.get_instance_state(instance_id=instance_id[0]) == "running"

        terminate_instance = EC2TerminateInstanceOperator(
            task_id="test_terminate_instance", instance_ids=instance_id
        )
        terminate_instance.execute(None)

        assert ec2_hook.get_instance_state(instance_id=instance_id[0]) == "terminated"

    @mock_aws
    def test_terminate_multiple_instances(self):
        ec2_hook = EC2Hook()
        create_instances = EC2CreateInstanceOperator(
            task_id="test_create_multiple_instances",
            image_id=self._get_image_id(hook=ec2_hook),
            min_count=5,
            max_count=5,
        )
        instance_ids = create_instances.execute(None)
        assert len(instance_ids) == 5

        for id in instance_ids:
            assert ec2_hook.get_instance_state(instance_id=id) == "running"

        terminate_instance = EC2TerminateInstanceOperator(
            task_id="test_terminate_instance", instance_ids=instance_ids
        )
        terminate_instance.execute(None)
        for id in instance_ids:
            assert ec2_hook.get_instance_state(instance_id=id) == "terminated"

    def test_template_fields(self):
        ec2_operator = EC2TerminateInstanceOperator(
            task_id="test_terminate_instance",
            instance_ids="test_image_id",
        )
        validate_template_fields(ec2_operator)


class TestEC2StartInstanceOperator(BaseEc2TestClass):
    def test_init(self):
        ec2_operator = EC2StartInstanceOperator(
            task_id="task_test",
            instance_id="i-123abc",
            aws_conn_id="aws_conn_test",
            region_name="region-test",
            check_interval=3,
        )
        assert ec2_operator.task_id == "task_test"
        assert ec2_operator.instance_id == "i-123abc"
        assert ec2_operator.aws_conn_id == "aws_conn_test"
        assert ec2_operator.region_name == "region-test"
        assert ec2_operator.check_interval == 3

    @mock_aws
    def test_start_instance(self):
        # create instance
        ec2_hook = EC2Hook()
        create_instance = EC2CreateInstanceOperator(
            image_id=self._get_image_id(ec2_hook),
            task_id="test_create_instance",
        )
        instance_id = create_instance.execute(None)

        # start instance
        start_test = EC2StartInstanceOperator(
            task_id="start_test",
            instance_id=instance_id[0],
        )
        start_test.execute(None)
        # assert instance state is running
        assert ec2_hook.get_instance_state(instance_id=instance_id[0]) == "running"

    def test_template_fields(self):
        ec2_operator = EC2StartInstanceOperator(
            task_id="task_test",
            instance_id="i-123abc",
            aws_conn_id="aws_conn_test",
            region_name="region-test",
            check_interval=3,
        )

        validate_template_fields(ec2_operator)


class TestEC2StopInstanceOperator(BaseEc2TestClass):
    def test_init(self):
        ec2_operator = EC2StopInstanceOperator(
            task_id="task_test",
            instance_id="i-123abc",
            aws_conn_id="aws_conn_test",
            region_name="region-test",
            check_interval=3,
        )
        assert ec2_operator.task_id == "task_test"
        assert ec2_operator.instance_id == "i-123abc"
        assert ec2_operator.aws_conn_id == "aws_conn_test"
        assert ec2_operator.region_name == "region-test"
        assert ec2_operator.check_interval == 3

    @mock_aws
    def test_stop_instance(self):
        # create instance
        ec2_hook = EC2Hook()
        create_instance = EC2CreateInstanceOperator(
            image_id=self._get_image_id(ec2_hook),
            task_id="test_create_instance",
        )
        instance_id = create_instance.execute(None)

        # stop instance
        stop_test = EC2StopInstanceOperator(
            task_id="stop_test",
            instance_id=instance_id[0],
        )
        stop_test.execute(None)
        # assert instance state is running
        assert ec2_hook.get_instance_state(instance_id=instance_id[0]) == "stopped"

    def test_template_fields(self):
        ec2_operator = EC2StopInstanceOperator(
            task_id="task_test",
            instance_id="i-123abc",
            aws_conn_id="aws_conn_test",
            region_name="region-test",
            check_interval=3,
        )

        validate_template_fields(ec2_operator)


class TestEC2HibernateInstanceOperator(BaseEc2TestClass):
    def test_init(self):
        ec2_operator = EC2HibernateInstanceOperator(
            task_id="task_test",
            instance_ids="i-123abc",
        )
        assert ec2_operator.task_id == "task_test"
        assert ec2_operator.instance_ids == "i-123abc"

    @mock_aws
    def test_hibernate_instance(self):
        # create instance
        ec2_hook = EC2Hook()
        create_instance = EC2CreateInstanceOperator(
            image_id=self._get_image_id(ec2_hook),
            task_id="test_create_instance",
            config={"HibernationOptions": {"Configured": True}},
        )
        instance_id = create_instance.execute(None)

        # hibernate instance
        hibernate_test = EC2HibernateInstanceOperator(
            task_id="hibernate_test",
            instance_ids=instance_id[0],
        )
        hibernate_test.execute(None)
        # assert instance state is stopped
        assert ec2_hook.get_instance_state(instance_id=instance_id[0]) == "stopped"

    @mock_aws
    def test_hibernate_multiple_instances(self):
        ec2_hook = EC2Hook()
        create_instances = EC2CreateInstanceOperator(
            task_id="test_create_multiple_instances",
            image_id=self._get_image_id(hook=ec2_hook),
            config={"HibernationOptions": {"Configured": True}},
            min_count=5,
            max_count=5,
        )
        instance_ids = create_instances.execute(None)
        assert len(instance_ids) == 5

        for id in instance_ids:
            assert ec2_hook.get_instance_state(instance_id=id) == "running"

        hibernate_instance = EC2HibernateInstanceOperator(
            task_id="test_hibernate_instance", instance_ids=instance_ids
        )
        hibernate_instance.execute(None)
        for id in instance_ids:
            assert ec2_hook.get_instance_state(instance_id=id) == "stopped"

    @mock_aws
    def test_cannot_hibernate_instance(self):
        # create instance
        ec2_hook = EC2Hook()
        create_instance = EC2CreateInstanceOperator(
            image_id=self._get_image_id(ec2_hook),
            task_id="test_create_instance",
        )
        instance_id = create_instance.execute(None)

        # hibernate instance
        hibernate_test = EC2HibernateInstanceOperator(
            task_id="hibernate_test",
            instance_ids=instance_id[0],
        )

        # assert hibernating an instance not configured for hibernation raises an error
        with pytest.raises(
            AirflowException,
            match="Instance .* is not configured for hibernation",
        ):
            hibernate_test.execute(None)

        # assert instance state is running
        assert ec2_hook.get_instance_state(instance_id=instance_id[0]) == "running"

    @mock_aws
    def test_cannot_hibernate_some_instances(self):
        # create instance
        ec2_hook = EC2Hook()
        create_instance_hibernate = EC2CreateInstanceOperator(
            image_id=self._get_image_id(ec2_hook),
            task_id="test_create_instance",
            config={"HibernationOptions": {"Configured": True}},
        )
        instance_id_hibernate = create_instance_hibernate.execute(None)
        create_instance_cannot_hibernate = EC2CreateInstanceOperator(
            image_id=self._get_image_id(ec2_hook),
            task_id="test_create_instance",
        )
        instance_id_cannot_hibernate = create_instance_cannot_hibernate.execute(None)
        instance_ids = [instance_id_hibernate[0], instance_id_cannot_hibernate[0]]

        # hibernate instance
        hibernate_test = EC2HibernateInstanceOperator(
            task_id="hibernate_test",
            instance_ids=instance_ids,
        )
        # assert hibernating an instance not configured for hibernation raises an error
        with pytest.raises(
            AirflowException,
            match="Instance .* is not configured for hibernation",
        ):
            hibernate_test.execute(None)

        # assert instance state is running
        for id in instance_ids:
            assert ec2_hook.get_instance_state(instance_id=id) == "running"

    def test_template_fields(self):
        ec2_operator = EC2HibernateInstanceOperator(
            task_id="task_test",
            instance_ids="i-123abc",
        )
        validate_template_fields(ec2_operator)


class TestEC2RebootInstanceOperator(BaseEc2TestClass):
    def test_init(self):
        ec2_operator = EC2RebootInstanceOperator(
            task_id="task_test",
            instance_ids="i-123abc",
        )
        assert ec2_operator.task_id == "task_test"
        assert ec2_operator.instance_ids == "i-123abc"

    @mock_aws
    def test_reboot_instance(self):
        # create instance
        ec2_hook = EC2Hook()
        create_instance = EC2CreateInstanceOperator(
            image_id=self._get_image_id(ec2_hook),
            task_id="test_create_instance",
        )
        instance_id = create_instance.execute(None)

        # reboot instance
        reboot_test = EC2RebootInstanceOperator(
            task_id="reboot_test",
            instance_ids=instance_id[0],
        )
        reboot_test.execute(None)
        # assert instance state is running
        assert ec2_hook.get_instance_state(instance_id=instance_id[0]) == "running"

    @mock_aws
    def test_reboot_multiple_instances(self):
        ec2_hook = EC2Hook()
        create_instances = EC2CreateInstanceOperator(
            task_id="test_create_multiple_instances",
            image_id=self._get_image_id(hook=ec2_hook),
            min_count=5,
            max_count=5,
        )
        instance_ids = create_instances.execute(None)
        assert len(instance_ids) == 5

        for id in instance_ids:
            assert ec2_hook.get_instance_state(instance_id=id) == "running"

        terminate_instance = EC2RebootInstanceOperator(
            task_id="test_reboot_instance", instance_ids=instance_ids
        )
        terminate_instance.execute(None)
        for id in instance_ids:
            assert ec2_hook.get_instance_state(instance_id=id) == "running"

    def test_template_fields(self):
        ec2_operator = EC2RebootInstanceOperator(
            task_id="task_test",
            instance_ids="i-123abc",
        )
        validate_template_fields(ec2_operator)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest
from dateutil.tz import tzlocal

from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.sensors.emr import EmrJobFlowSensor
from airflow.providers.amazon.aws.triggers.emr import EmrTerminateJobFlowTrigger
from airflow.providers.common.compat.sdk import AirflowException, TaskDeferred

DESCRIBE_CLUSTER_STARTING_RETURN = {
    "Cluster": {
        "Applications": [{"Name": "Spark", "Version": "1.6.1"}],
        "AutoTerminate": True,
        "Configurations": [],
        "Ec2InstanceAttributes": {"IamInstanceProfile": "EMR_EC2_DefaultRole"},
        "Id": "j-27ZY9GBEEU2GU",
        "LogUri": "s3n://some-location/",
        "Name": "PiCalc",
        "NormalizedInstanceHours": 0,
        "ReleaseLabel": "emr-4.6.0",
        "ServiceRole": "EMR_DefaultRole",
        "Status": {
            "State": "STARTING",
            "StateChangeReason": {},
            "Timeline": {
                "CreationDateTime": datetime.datetime(2016, 6, 27, 21, 5, 2, 348000, tzinfo=tzlocal())
            },
        },
        "Tags": [{"Key": "app", "Value": "analytics"}, {"Key": "environment", "Value": "development"}],
        "TerminationProtected": False,
        "VisibleToAllUsers": True,
    },
    "ResponseMetadata": {"HTTPStatusCode": 200, "RequestId": "d5456308-3caa-11e6-9d46-951401f04e0e"},
}

DESCRIBE_CLUSTER_BOOTSTRAPPING_RETURN = {
    "Cluster": {
        "Applications": [{"Name": "Spark", "Version": "1.6.1"}],
        "AutoTerminate": True,
        "Configurations": [],
        "Ec2InstanceAttributes": {"IamInstanceProfile": "EMR_EC2_DefaultRole"},
        "Id": "j-27ZY9GBEEU2GU",
        "LogUri": "s3n://some-location/",
        "Name": "PiCalc",
        "NormalizedInstanceHours": 0,
        "ReleaseLabel": "emr-4.6.0",
        "ServiceRole": "EMR_DefaultRole",
        "Status": {
            "State": "BOOTSTRAPPING",
            "StateChangeReason": {},
            "Timeline": {
                "CreationDateTime": datetime.datetime(2016, 6, 27, 21, 5, 2, 348000, tzinfo=tzlocal())
            },
        },
        "Tags": [{"Key": "app", "Value": "analytics"}, {"Key": "environment", "Value": "development"}],
        "TerminationProtected": False,
        "VisibleToAllUsers": True,
    },
    "ResponseMetadata": {"HTTPStatusCode": 200, "RequestId": "d5456308-3caa-11e6-9d46-951401f04e0e"},
}

DESCRIBE_CLUSTER_RUNNING_RETURN = {
    "Cluster": {
        "Applications": [{"Name": "Spark", "Version": "1.6.1"}],
        "AutoTerminate": True,
        "Configurations": [],
        "Ec2InstanceAttributes": {"IamInstanceProfile": "EMR_EC2_DefaultRole"},
        "Id": "j-27ZY9GBEEU2GU",
        "LogUri": "s3n://some-location/",
        "Name": "PiCalc",
        "NormalizedInstanceHours": 0,
        "ReleaseLabel": "emr-4.6.0",
        "ServiceRole": "EMR_DefaultRole",
        "Status": {
            "State": "RUNNING",
            "StateChangeReason": {},
            "Timeline": {
                "CreationDateTime": datetime.datetime(2016, 6, 27, 21, 5, 2, 348000, tzinfo=tzlocal())
            },
        },
        "Tags": [{"Key": "app", "Value": "analytics"}, {"Key": "environment", "Value": "development"}],
        "TerminationProtected": False,
        "VisibleToAllUsers": True,
    },
    "ResponseMetadata": {"HTTPStatusCode": 200, "RequestId": "d5456308-3caa-11e6-9d46-951401f04e0e"},
}

DESCRIBE_CLUSTER_WAITING_RETURN = {
    "Cluster": {
        "Applications": [{"Name": "Spark", "Version": "1.6.1"}],
        "AutoTerminate": True,
        "Configurations": [],
        "Ec2InstanceAttributes": {"IamInstanceProfile": "EMR_EC2_DefaultRole"},
        "Id": "j-27ZY9GBEEU2GU",
        "LogUri": "s3n://some-location/",
        "Name": "PiCalc",
        "NormalizedInstanceHours": 0,
        "ReleaseLabel": "emr-4.6.0",
        "ServiceRole": "EMR_DefaultRole",
        "Status": {
            "State": "WAITING",
            "StateChangeReason": {},
            "Timeline": {
                "CreationDateTime": datetime.datetime(2016, 6, 27, 21, 5, 2, 348000, tzinfo=tzlocal())
            },
        },
        "Tags": [{"Key": "app", "Value": "analytics"}, {"Key": "environment", "Value": "development"}],
        "TerminationProtected": False,
        "VisibleToAllUsers": True,
    },
    "ResponseMetadata": {"HTTPStatusCode": 200, "RequestId": "d5456308-3caa-11e6-9d46-951401f04e0e"},
}

DESCRIBE_CLUSTER_TERMINATED_RETURN = {
    "Cluster": {
        "Applications": [{"Name": "Spark", "Version": "1.6.1"}],
        "AutoTerminate": True,
        "Configurations": [],
        "Ec2InstanceAttributes": {"IamInstanceProfile": "EMR_EC2_DefaultRole"},
        "Id": "j-27ZY9GBEEU2GU",
        "LogUri": "s3n://some-location/",
        "Name": "PiCalc",
        "NormalizedInstanceHours": 0,
        "ReleaseLabel": "emr-4.6.0",
        "ServiceRole": "EMR_DefaultRole",
        "Status": {
            "State": "TERMINATED",
            "StateChangeReason": {},
            "Timeline": {
                "CreationDateTime": datetime.datetime(2016, 6, 27, 21, 5, 2, 348000, tzinfo=tzlocal())
            },
        },
        "Tags": [{"Key": "app", "Value": "analytics"}, {"Key": "environment", "Value": "development"}],
        "TerminationProtected": False,
        "VisibleToAllUsers": True,
    },
    "ResponseMetadata": {"HTTPStatusCode": 200, "RequestId": "d5456308-3caa-11e6-9d46-951401f04e0e"},
}

DESCRIBE_CLUSTER_TERMINATED_WITH_ERRORS_RETURN = {
    "Cluster": {
        "Applications": [{"Name": "Spark", "Version": "1.6.1"}],
        "AutoTerminate": True,
        "Configurations": [],
        "Ec2InstanceAttributes": {"IamInstanceProfile": "EMR_EC2_DefaultRole"},
        "Id": "j-27ZY9GBEEU2GU",
        "LogUri": "s3n://some-location/",
        "Name": "PiCalc",
        "NormalizedInstanceHours": 0,
        "ReleaseLabel": "emr-4.6.0",
        "ServiceRole": "EMR_DefaultRole",
        "Status": {
            "State": "TERMINATED_WITH_ERRORS",
            "StateChangeReason": {
                "Code": "BOOTSTRAP_FAILURE",
                "Message": "Master instance (i-0663047709b12345c) failed attempting to "
                "download bootstrap action 1 file from S3",
            },
            "Timeline": {
                "CreationDateTime": datetime.datetime(2016, 6, 27, 21, 5, 2, 348000, tzinfo=tzlocal())
            },
        },
        "Tags": [{"Key": "app", "Value": "analytics"}, {"Key": "environment", "Value": "development"}],
        "TerminationProtected": False,
        "VisibleToAllUsers": True,
    },
    "ResponseMetadata": {"HTTPStatusCode": 200, "RequestId": "d5456308-3caa-11e6-9d46-951401f04e0e"},
}


@pytest.fixture
def mocked_hook_client():
    with mock.patch("airflow.providers.amazon.aws.hooks.emr.EmrHook.conn") as m:
        yield m


class TestEmrJobFlowSensor:
    def test_execute_calls_with_the_job_flow_id_until_it_reaches_a_target_state(self, mocked_hook_client):
        mocked_hook_client.describe_cluster.side_effect = [
            DESCRIBE_CLUSTER_STARTING_RETURN,
            DESCRIBE_CLUSTER_RUNNING_RETURN,
            DESCRIBE_CLUSTER_TERMINATED_RETURN,
        ]

        operator = EmrJobFlowSensor(
            task_id="test_task", poke_interval=0, job_flow_id="j-8989898989", aws_conn_id="aws_default"
        )
        with patch.object(S3Hook, "parse_s3_url", return_value="valid_uri"):
            operator.execute(MagicMock())

        assert mocked_hook_client.describe_cluster.call_count == 3
        # make sure it was called with the job_flow_id
        calls = [mock.call(ClusterId="j-8989898989")] * 3
        mocked_hook_client.describe_cluster.assert_has_calls(calls)

    def test_execute_calls_with_the_job_flow_id_until_it_reaches_failed_state_with_exception(
        self, mocked_hook_client
    ):
        mocked_hook_client.describe_cluster.side_effect = [
            DESCRIBE_CLUSTER_RUNNING_RETURN,
            DESCRIBE_CLUSTER_TERMINATED_WITH_ERRORS_RETURN,
        ]

        operator = EmrJobFlowSensor(
            task_id="test_task", poke_interval=0, job_flow_id="j-8989898989", aws_conn_id="aws_default"
        )
        with pytest.raises(AirflowException):
            operator.execute(MagicMock())

        # make sure we called twice
        assert mocked_hook_client.describe_cluster.call_count == 2
        # make sure it was called with the job_flow_id
        calls = [mock.call(ClusterId="j-8989898989")] * 2
        mocked_hook_client.describe_cluster.assert_has_calls(calls=calls)

    def test_different_target_states(self, mocked_hook_client):
        mocked_hook_client.describe_cluster.side_effect = [
            DESCRIBE_CLUSTER_STARTING_RETURN,  # return False
            DESCRIBE_CLUSTER_BOOTSTRAPPING_RETURN,  # return False
            DESCRIBE_CLUSTER_RUNNING_RETURN,  # return True
            DESCRIBE_CLUSTER_WAITING_RETURN,  # will not be used
            DESCRIBE_CLUSTER_TERMINATED_RETURN,  # will not be used
            DESCRIBE_CLUSTER_TERMINATED_WITH_ERRORS_RETURN,  # will not be used
        ]

        operator = EmrJobFlowSensor(
            task_id="test_task",
            poke_interval=0,
            job_flow_id="j-8989898989",
            aws_conn_id="aws_default",
            target_states=["RUNNING", "WAITING"],
        )

        operator.execute(MagicMock())

        assert mocked_hook_client.describe_cluster.call_count == 3
        # make sure it was called with the job_flow_id
        calls = [mock.call(ClusterId="j-8989898989")] * 3
        mocked_hook_client.describe_cluster.assert_has_calls(calls)

    def test_sensor_defer(self):
        """Test the execute method raise TaskDeferred if running sensor in deferrable mode"""
        sensor = EmrJobFlowSensor(
            task_id="test_task",
            poke_interval=0,
            job_flow_id="j-8989898989",
            aws_conn_id="aws_default",
            target_states=["RUNNING", "WAITING"],
            deferrable=True,
        )

        with patch.object(EmrJobFlowSensor, "poke", return_value=False):
            with pytest.raises(TaskDeferred) as exc:
                sensor.execute(context=None)

        assert isinstance(exc.value.trigger, EmrTerminateJobFlowTrigger), (
            f"{exc.value.trigger} is not a EmrTerminateJobFlowTrigger"
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import boto3
import pytest
from moto import mock_aws

from airflow.models import DAG
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.transfers.sftp_to_s3 import SFTPToS3Operator
from airflow.providers.ssh.hooks.ssh import SSHHook
from airflow.providers.ssh.operators.ssh import SSHOperator
from airflow.utils.timezone import datetime

from tests_common.test_utils.config import conf_vars

pytestmark = pytest.mark.db_test

BUCKET = "test-bucket"
S3_KEY = "test/test_1_file.csv"
SFTP_PATH = "/tmp/remote_path.txt"
SFTP_CONN_ID = "ssh_default"
S3_CONN_ID = "aws_default"

SFTP_MOCK_FILE = "test_sftp_file.csv"
S3_MOCK_FILES = "test_1_file.csv"

TEST_DAG_ID = "unit_tests_sftp_tos3_op"
DEFAULT_DATE = datetime(2018, 1, 1)


class TestSFTPToS3Operator:
    def setup_method(self):
        hook = SSHHook(ssh_conn_id="ssh_default")

        hook.no_host_key_check = True
        dag = DAG(
            f"{TEST_DAG_ID}test_schedule_dag_once",
            schedule="@once",
            start_date=DEFAULT_DATE,
        )

        self.hook = hook

        self.ssh_client = self.hook.get_conn()
        self.sftp_client = self.ssh_client.open_sftp()

        self.dag = dag
        self.s3_bucket = BUCKET
        self.sftp_path = SFTP_PATH
        self.s3_key = S3_KEY

    @pytest.mark.parametrize("use_temp_file", [True, False])
    @mock_aws
    @conf_vars({("core", "enable_xcom_pickling"): "True"})
    def test_sftp_to_s3_operation(self, use_temp_file):
        # Setting
        test_remote_file_content = (
            "This is remote file content \n which is also multiline "
            "another line here \n this is last line. EOF"
        )

        # create a test file remotely
        create_file_task = SSHOperator(
            task_id="test_create_file",
            ssh_hook=self.hook,
            command=f"echo '{test_remote_file_content}' > {self.sftp_path}",
            do_xcom_push=True,
            dag=self.dag,
        )
        assert create_file_task is not None
        create_file_task.execute(None)

        # Test for creation of s3 bucket
        s3_hook = S3Hook(aws_conn_id=None)
        conn = boto3.client("s3")
        conn.create_bucket(Bucket=self.s3_bucket)
        assert s3_hook.check_for_bucket(self.s3_bucket)

        # get remote file to local
        run_task = SFTPToS3Operator(
            s3_bucket=BUCKET,
            s3_key=S3_KEY,
            sftp_path=SFTP_PATH,
            sftp_conn_id=SFTP_CONN_ID,
            s3_conn_id=S3_CONN_ID,
            use_temp_file=use_temp_file,
            task_id="test_sftp_to_s3",
            dag=self.dag,
        )
        assert run_task is not None

        run_task.execute(None)

        # Check if object was created in s3
        objects_in_dest_bucket = conn.list_objects(Bucket=self.s3_bucket, Prefix=self.s3_key)
        # there should be object found, and there should only be one object found
        assert len(objects_in_dest_bucket["Contents"]) == 1

        # the object found should be consistent with dest_key specified earlier
        assert objects_in_dest_bucket["Contents"][0]["Key"] == self.s3_key

        # Clean up after finishing with test
        conn.delete_object(Bucket=self.s3_bucket, Key=self.s3_key)
        conn.delete_bucket(Bucket=self.s3_bucket)
        assert not s3_hook.check_for_bucket(self.s3_bucket)

    @pytest.mark.parametrize("fail_on_file_not_exist", [True, False])
    @mock_aws
    @conf_vars({("core", "enable_xcom_pickling"): "True"})
    def test_sftp_to_s3_fail_on_file_not_exist(self, fail_on_file_not_exist):
        s3_hook = S3Hook(aws_conn_id=None)
        conn = boto3.client("s3")
        conn.create_bucket(Bucket=self.s3_bucket)
        assert s3_hook.check_for_bucket(self.s3_bucket)

        if fail_on_file_not_exist:
            with pytest.raises(FileNotFoundError):
                SFTPToS3Operator(
                    s3_bucket=self.s3_bucket,
                    s3_key=self.s3_key,
                    sftp_path="/tmp/wrong_path.txt",
                    sftp_conn_id=SFTP_CONN_ID,
                    s3_conn_id=S3_CONN_ID,
                    fail_on_file_not_exist=fail_on_file_not_exist,
                    task_id="test_sftp_to_s3",
                    dag=self.dag,
                ).execute(None)
        else:
            SFTPToS3Operator(
                s3_bucket=self.s3_bucket,
                s3_key=self.s3_key,
                sftp_path=self.sftp_path,
                sftp_conn_id=SFTP_CONN_ID,
                s3_conn_id=S3_CONN_ID,
                fail_on_file_not_exist=fail_on_file_not_exist,
                task_id="test_sftp_to_s3",
                dag=self.dag,
            ).execute(None)

        conn.delete_object(Bucket=self.s3_bucket, Key=self.s3_key)
        conn.delete_bucket(Bucket=self.s3_bucket)
        assert not s3_hook.check_for_bucket(self.s3_bucket)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, PropertyMock, patch

from airflow.providers.amazon.aws.hooks.emr import EmrServerlessHook

task_id = "test_emr_serverless_create_application_operator"
application_id = "test_application_id"
release_label = "test"
job_type = "test"
client_request_token = "eac427d0-1c6d4df=-96aa-32423412"
config = {"name": "test_application_emr_serverless"}


class TestEmrServerlessHook:
    def test_conn_attribute(self):
        hook = EmrServerlessHook(aws_conn_id="aws_default")
        assert hasattr(hook, "conn")
        # Testing conn is a cached property
        conn = hook.conn
        conn2 = hook.conn
        assert conn is conn2

    @patch.object(EmrServerlessHook, "conn", new_callable=PropertyMock)
    def test_cancel_jobs(self, conn_mock: MagicMock):
        conn_mock().get_paginator().paginate.return_value = [{"jobRuns": [{"id": "job1"}, {"id": "job2"}]}]
        hook = EmrServerlessHook(aws_conn_id="aws_default")
        waiter_mock = MagicMock()
        hook.get_waiter = waiter_mock  # type:ignore[method-assign]

        hook.cancel_running_jobs("app")

        assert conn_mock().cancel_job_run.call_count == 2
        conn_mock().cancel_job_run.assert_any_call(applicationId="app", jobRunId="job1")
        conn_mock().cancel_job_run.assert_any_call(applicationId="app", jobRunId="job2")
        waiter_mock.assert_called_with("no_job_running")

    @patch.object(EmrServerlessHook, "conn", new_callable=PropertyMock)
    def test_cancel_jobs_several_calls(self, conn_mock: MagicMock):
        conn_mock().get_paginator().paginate.return_value = [
            {"jobRuns": [{"id": "job1"}, {"id": "job2"}]},
            {"jobRuns": [{"id": "job3"}, {"id": "job4"}]},
        ]
        hook = EmrServerlessHook(aws_conn_id="aws_default")
        waiter_mock = MagicMock()
        hook.get_waiter = waiter_mock  # type:ignore[method-assign]

        hook.cancel_running_jobs("app")

        assert conn_mock().cancel_job_run.call_count == 4
        waiter_mock.assert_called_once()  # we should wait once for all jobs, not once per page

    @patch.object(EmrServerlessHook, "conn", new_callable=PropertyMock)
    def test_cancel_jobs_but_no_jobs(self, conn_mock: MagicMock):
        conn_mock.return_value.list_job_runs.return_value = {"jobRuns": []}
        hook = EmrServerlessHook(aws_conn_id="aws_default")

        hook.cancel_running_jobs("app")

        # nothing very interesting should happen
        conn_mock.assert_called_once()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook
from airflow.providers.common.compat.sdk import AirflowException


class ComprehendHook(AwsBaseHook):
    """
    Interact with AWS Comprehend.

    Provide thin wrapper around :external+boto3:py:class:`boto3.client("comprehend") <Comprehend.Client>`.

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    """

    def __init__(self, *args, **kwargs) -> None:
        kwargs["client_type"] = "comprehend"
        super().__init__(*args, **kwargs)

    def validate_document_classifier_training_status(
        self, document_classifier_arn: str, fail_on_warnings: bool = False
    ) -> None:
        """
        Log the Information about the document classifier.

        NumberOfLabels
        NumberOfTrainedDocuments
        NumberOfTestDocuments
        EvaluationMetrics

        """
        response = self.conn.describe_document_classifier(DocumentClassifierArn=document_classifier_arn)

        status = response["DocumentClassifierProperties"]["Status"]

        if status == "TRAINED_WITH_WARNING":
            self.log.info(
                "AWS Comprehend document classifier training completed with %s, Message: %s please review the skipped files folder in the output location %s",
                status,
                response["DocumentClassifierProperties"]["Message"],
                response["DocumentClassifierProperties"]["OutputDataConfig"]["S3Uri"],
            )

            if fail_on_warnings:
                raise AirflowException("Warnings in AWS Comprehend document classifier training.")

        self.log.info(
            "AWS Comprehend document classifier metadata: %s",
            response["DocumentClassifierProperties"]["ClassifierMetadata"],
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow.providers.amazon.aws.hooks.neptune import NeptuneHook
from airflow.providers.amazon.aws.operators.neptune import (
    NeptuneStartDbClusterOperator,
    NeptuneStopDbClusterOperator,
)

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]


from system.amazon.aws.utils import SystemTestContextBuilder

DAG_ID = "example_neptune"

sys_test_context_task = SystemTestContextBuilder().build()


@task
def create_cluster(cluster_id):
    hook = NeptuneHook()
    hook.conn.create_db_cluster(DBClusterIdentifier=cluster_id, Engine="neptune", DeletionProtection=False)
    hook.wait_for_cluster_availability(cluster_id=cluster_id)


@task
def delete_cluster(cluster_id):
    hook = NeptuneHook()
    hook.conn.delete_db_cluster(DBClusterIdentifier=cluster_id, SkipFinalSnapshot=True)


with DAG(
    dag_id=DAG_ID,
    start_date=datetime(2021, 1, 1),
    schedule="@once",
    catchup=False,
) as dag:
    test_context = sys_test_context_task()

    env_id = test_context["ENV_ID"]
    cluster_id = f"{env_id}-cluster"

    # [START howto_operator_start_neptune_cluster]
    start_cluster = NeptuneStartDbClusterOperator(task_id="start_task", db_cluster_id=cluster_id)
    # [END howto_operator_start_neptune_cluster]

    # [START howto_operator_stop_neptune_cluster]
    stop_cluster = NeptuneStopDbClusterOperator(task_id="stop_task", db_cluster_id=cluster_id)
    # [END howto_operator_stop_neptune_cluster]

    chain(
        # TEST SETUP
        test_context,
        create_cluster(cluster_id),
        # TEST BODY
        start_cluster,
        stop_cluster,
        # TEST TEARDOWN
        delete_cluster(cluster_id),
    )
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.amazon.aws.auth_manager.user import AwsAuthManagerUser


@pytest.fixture
def user():
    return AwsAuthManagerUser(user_id="user_id", groups=[])


class TestAwsAuthManagerUser:
    def test_get_id(self, user):
        assert user.get_id() == "user_id"

    def test_get_name_with_username(self, user):
        user.username = "username"
        assert user.get_name() == "username"

    def test_get_name_with_email(self, user):
        user.email = "email"
        assert user.get_name() == "email"

    def test_get_name_with_user_id(self, user):
        user.user_id = "user_id"
        assert user.get_name() == "user_id"

    def test_get_groups(self, user):
        assert user.get_groups() == []



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.amazon.aws.hooks.comprehend import ComprehendHook
from airflow.providers.amazon.aws.sensors.comprehend import (
    ComprehendCreateDocumentClassifierCompletedSensor,
    ComprehendStartPiiEntitiesDetectionJobCompletedSensor,
)
from airflow.providers.common.compat.sdk import AirflowException


class TestComprehendStartPiiEntitiesDetectionJobCompletedSensor:
    SENSOR = ComprehendStartPiiEntitiesDetectionJobCompletedSensor

    def setup_method(self):
        self.default_op_kwargs = dict(
            task_id="test_pii_entities_detection_job_sensor",
            job_id="job_id",
            poke_interval=5,
            max_retries=1,
        )
        self.sensor = self.SENSOR(**self.default_op_kwargs, aws_conn_id=None)

    def test_base_aws_op_attributes(self):
        op = self.SENSOR(**self.default_op_kwargs)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

        op = self.SENSOR(
            **self.default_op_kwargs,
            aws_conn_id="aws-test-custom-conn",
            region_name="eu-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.aws_conn_id == "aws-test-custom-conn"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

    @pytest.mark.parametrize("state", SENSOR.SUCCESS_STATES)
    @mock.patch.object(ComprehendHook, "conn")
    def test_poke_success_state(self, mock_conn, state):
        mock_conn.describe_pii_entities_detection_job.return_value = {
            "PiiEntitiesDetectionJobProperties": {"JobStatus": state}
        }
        assert self.sensor.poke({}) is True

    @pytest.mark.parametrize("state", SENSOR.INTERMEDIATE_STATES)
    @mock.patch.object(ComprehendHook, "conn")
    def test_intermediate_state(self, mock_conn, state):
        mock_conn.describe_pii_entities_detection_job.return_value = {
            "PiiEntitiesDetectionJobProperties": {"JobStatus": state}
        }
        assert self.sensor.poke({}) is False

    @pytest.mark.parametrize("state", SENSOR.FAILURE_STATES)
    @mock.patch.object(ComprehendHook, "conn")
    def test_poke_failure_states(self, mock_conn, state):
        mock_conn.describe_pii_entities_detection_job.return_value = {
            "PiiEntitiesDetectionJobProperties": {"JobStatus": state}
        }
        sensor = self.SENSOR(**self.default_op_kwargs, aws_conn_id=None)

        with pytest.raises(AirflowException, match=sensor.FAILURE_MESSAGE):
            sensor.poke({})


class TestComprehendCreateDocumentClassifierCompletedSensor:
    SENSOR = ComprehendCreateDocumentClassifierCompletedSensor
    DOCUMENT_CLASSIFIER_ARN = (
        "arn:aws:comprehend:us-east-1:123456789012:document-classifier/insurance-classifier/version/v1"
    )
    EVALUATION_METRICS = {
        "EvaluationMetrics": {
            "Accuracy": 1,
            "Precision": 1,
            "Recall": 1,
            "F1Score": 1,
            "MicroPrecision": 1,
            "MicroRecall": 1,
            "MicroF1Score": 1,
            "HammingLoss": 0,
        }
    }

    def setup_method(self):
        self.default_op_kwargs = dict(
            task_id="test_create_document_classifier_sensor",
            document_classifier_arn=self.DOCUMENT_CLASSIFIER_ARN,
            fail_on_warnings=False,
            poke_interval=5,
            max_retries=1,
        )
        self.sensor = self.SENSOR(**self.default_op_kwargs, aws_conn_id=None)

    def test_base_aws_op_attributes(self):
        op = self.SENSOR(**self.default_op_kwargs)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

        op = self.SENSOR(
            **self.default_op_kwargs,
            aws_conn_id="aws-test-custom-conn",
            region_name="eu-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.aws_conn_id == "aws-test-custom-conn"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

    @pytest.mark.parametrize(
        ("state", "message", "output"),
        [
            pytest.param("TRAINED", "", "s3://test-output", id="training succeeded"),
            pytest.param(
                "TRAINED_WITH_WARNING",
                "Unable to parse some documents. See details in the output S3 location",
                "s3://test-output",
                id="trained with warning",
            ),
        ],
    )
    @mock.patch.object(ComprehendHook, "conn")
    def test_poke_success_state(self, mock_conn, state, message, output):
        mock_conn.describe_document_classifier.return_value = {
            "DocumentClassifierProperties": {
                "Status": state,
                "Message": message,
                "OutputDataConfig": {"S3Uri": output},
                "ClassifierMetadata": self.EVALUATION_METRICS,
            }
        }

        assert self.sensor.poke({}) is True

    @pytest.mark.parametrize("state", SENSOR.INTERMEDIATE_STATES)
    @mock.patch.object(ComprehendHook, "conn")
    def test_intermediate_state(self, mock_conn, state):
        mock_conn.describe_document_classifier.return_value = {
            "DocumentClassifierProperties": {"Status": state}
        }
        assert self.sensor.poke({}) is False

    @pytest.mark.parametrize("state", SENSOR.FAILURE_STATES)
    @mock.patch.object(ComprehendHook, "conn")
    def test_poke_failure_states(self, mock_conn, state):
        mock_conn.describe_document_classifier.return_value = {
            "DocumentClassifierProperties": {"Status": state}
        }
        sensor = self.SENSOR(**self.default_op_kwargs, aws_conn_id=None)

        with pytest.raises(AirflowException, match=sensor.FAILURE_MESSAGE):
            sensor.poke({})



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

import boto3

from airflow.providers.amazon.aws.hooks.ecs import EcsClusterStates
from airflow.providers.amazon.aws.operators.ecs import (
    EcsCreateClusterOperator,
    EcsDeleteClusterOperator,
    EcsDeregisterTaskDefinitionOperator,
    EcsRegisterTaskDefinitionOperator,
    EcsRunTaskOperator,
)
from airflow.providers.amazon.aws.sensors.ecs import (
    EcsClusterStateSensor,
    EcsTaskDefinitionStateSensor,
)

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

DAG_ID = "example_ecs"

# Externally fetched variables:
EXISTING_CLUSTER_NAME_KEY = "CLUSTER_NAME"
EXISTING_CLUSTER_SUBNETS_KEY = "SUBNETS"

sys_test_context_task = (
    SystemTestContextBuilder()
    # NOTE:  Creating a functional ECS Cluster which uses EC2 requires manually creating
    # and configuring a number of resources such as autoscaling groups, networking
    # etc. which is out of scope for this demo and time-consuming for a system test
    # To simplify this demo and make it run in a reasonable length of time as a
    # system test, follow the steps below to create a new cluster on the AWS Console
    # which handles all asset creation and configuration using default values:
    # 1. https://us-east-1.console.aws.amazon.com/ecs/home?region=us-east-1#/clusters
    # 2. Select "EC2 Linux + Networking" and hit "Next"
    # 3. Name your cluster in the first field and click Create
    .add_variable(EXISTING_CLUSTER_NAME_KEY)
    .add_variable(EXISTING_CLUSTER_SUBNETS_KEY, split_string=True)
    .build()
)


@task
def get_region():
    return boto3.session.Session().region_name


@task(trigger_rule=TriggerRule.ALL_DONE)
def clean_logs(group_name: str):
    client = boto3.client("logs")
    client.delete_log_group(logGroupName=group_name)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]
    existing_cluster_name = test_context[EXISTING_CLUSTER_NAME_KEY]
    existing_cluster_subnets = test_context[EXISTING_CLUSTER_SUBNETS_KEY]

    new_cluster_name = f"{env_id}-cluster"
    container_name = f"{env_id}-container"
    family_name = f"{env_id}-task-definition"
    asg_name = f"{env_id}-asg"

    aws_region = get_region()
    log_group_name = f"/ecs_test/{env_id}"

    # [START howto_operator_ecs_create_cluster]
    create_cluster = EcsCreateClusterOperator(
        task_id="create_cluster",
        cluster_name=new_cluster_name,
    )
    # [END howto_operator_ecs_create_cluster]

    # EcsCreateClusterOperator waits by default, setting as False to test the Sensor below.
    create_cluster.wait_for_completion = False

    # [START howto_sensor_ecs_cluster_state]
    await_cluster = EcsClusterStateSensor(
        task_id="await_cluster",
        cluster_name=new_cluster_name,
    )
    # [END howto_sensor_ecs_cluster_state]

    # [START howto_operator_ecs_register_task_definition]
    register_task = EcsRegisterTaskDefinitionOperator(
        task_id="register_task",
        family=family_name,
        container_definitions=[
            {
                "name": container_name,
                "image": "ubuntu",
                "workingDirectory": "/usr/bin",
                "entryPoint": ["sh", "-c"],
                "command": ["ls"],
                "logConfiguration": {
                    "logDriver": "awslogs",
                    "options": {
                        "awslogs-group": log_group_name,
                        "awslogs-region": aws_region,
                        "awslogs-create-group": "true",
                        "awslogs-stream-prefix": "ecs",
                    },
                },
            },
        ],
        register_task_kwargs={
            "cpu": "256",
            "memory": "512",
            "networkMode": "awsvpc",
        },
    )
    # [END howto_operator_ecs_register_task_definition]

    # [START howto_sensor_ecs_task_definition_state]
    await_task_definition = EcsTaskDefinitionStateSensor(
        task_id="await_task_definition",
        task_definition=register_task.output,
    )
    # [END howto_sensor_ecs_task_definition_state]

    # [START howto_operator_ecs_run_task]
    run_task = EcsRunTaskOperator(
        task_id="run_task",
        reattach=True,
        cluster=existing_cluster_name,
        task_definition=register_task.output,
        overrides={
            "containerOverrides": [
                {
                    "name": container_name,
                    "command": ["echo hello world"],
                },
            ],
        },
        network_configuration={"awsvpcConfiguration": {"subnets": existing_cluster_subnets}},
        # [START howto_awslogs_ecs]
        awslogs_group=log_group_name,
        awslogs_region=aws_region,
        awslogs_stream_prefix=f"ecs/{container_name}",
        # [END howto_awslogs_ecs]
    )
    # [END howto_operator_ecs_run_task]
    # The default is 6 seconds between checks, which is very aggressive, setting to 60s to reduce throttling errors.
    run_task.waiter_delay = 60

    # [START howto_operator_ecs_deregister_task_definition]
    deregister_task = EcsDeregisterTaskDefinitionOperator(
        task_id="deregister_task",
        task_definition=register_task.output,
    )
    # [END howto_operator_ecs_deregister_task_definition]
    deregister_task.trigger_rule = TriggerRule.ALL_DONE

    # [START howto_operator_ecs_delete_cluster]
    delete_cluster = EcsDeleteClusterOperator(
        task_id="delete_cluster",
        cluster_name=new_cluster_name,
    )
    # [END howto_operator_ecs_delete_cluster]
    delete_cluster.trigger_rule = TriggerRule.ALL_DONE

    # EcsDeleteClusterOperator waits by default, setting as False to test the Sensor below.
    delete_cluster.wait_for_completion = False

    # [START howto_operator_ecs_delete_cluster]
    await_delete_cluster = EcsClusterStateSensor(
        task_id="await_delete_cluster",
        cluster_name=new_cluster_name,
        target_state=EcsClusterStates.INACTIVE,
    )
    # [END howto_operator_ecs_delete_cluster]

    chain(
        # TEST SETUP
        test_context,
        aws_region,
        # TEST BODY
        create_cluster,
        await_cluster,
        register_task,
        await_task_definition,
        run_task,
        deregister_task,
        delete_cluster,
        await_delete_cluster,
        clean_logs(log_group_name),
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import boto3
import botocore.exceptions
import pytest
from botocore.exceptions import ClientError
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.glue_catalog import GlueCatalogHook
from airflow.providers.common.compat.sdk import AirflowException

DB_NAME = "db"
TABLE_NAME = "table"
TABLE_INPUT = {
    "Name": TABLE_NAME,
    "StorageDescriptor": {
        "Columns": [{"Name": "string", "Type": "string", "Comment": "string"}],
        "Location": f"s3://mybucket/{DB_NAME}/{TABLE_NAME}",
    },
}
PARTITION_INPUT: dict = {
    "Values": [],
}


@mock_aws
class TestGlueCatalogHook:
    def setup_method(self, method):
        self.client = boto3.client("glue", region_name="us-east-1")
        self.hook = GlueCatalogHook(region_name="us-east-1")

    def test_get_conn_returns_a_boto3_connection(self):
        hook = GlueCatalogHook(region_name="us-east-1")
        assert hook.get_conn() is not None

    def test_conn_id(self):
        hook = GlueCatalogHook(aws_conn_id="my_aws_conn_id", region_name="us-east-1")
        assert hook.aws_conn_id == "my_aws_conn_id"

    def test_region(self):
        hook = GlueCatalogHook(region_name="us-west-2")
        assert hook.region_name == "us-west-2"

    @mock.patch.object(GlueCatalogHook, "get_conn")
    def test_get_partitions_empty(self, mock_get_conn):
        response = set()
        mock_get_conn.get_paginator.paginate.return_value = response
        hook = GlueCatalogHook(region_name="us-east-1")

        assert hook.get_partitions("db", "tbl") == set()

    @mock.patch.object(GlueCatalogHook, "get_conn")
    def test_get_partitions(self, mock_get_conn):
        response = [{"Partitions": [{"Values": ["2015-01-01"]}]}]
        mock_paginator = mock.Mock()
        mock_paginator.paginate.return_value = response
        mock_conn = mock.Mock()
        mock_conn.get_paginator.return_value = mock_paginator
        mock_get_conn.return_value = mock_conn
        hook = GlueCatalogHook(region_name="us-east-1")
        result = hook.get_partitions("db", "tbl", expression="foo=bar", page_size=2, max_items=3)

        assert result == {("2015-01-01",)}
        mock_conn.get_paginator.assert_called_once_with("get_partitions")
        mock_paginator.paginate.assert_called_once_with(
            DatabaseName="db",
            TableName="tbl",
            Expression="foo=bar",
            PaginationConfig={"PageSize": 2, "MaxItems": 3},
        )

    @mock.patch.object(GlueCatalogHook, "get_partitions")
    def test_check_for_partition(self, mock_get_partitions):
        mock_get_partitions.return_value = {("2018-01-01",)}
        hook = GlueCatalogHook(region_name="us-east-1")

        assert hook.check_for_partition("db", "tbl", "expr")
        mock_get_partitions.assert_called_once_with("db", "tbl", "expr", max_items=1)

    @mock.patch.object(GlueCatalogHook, "get_partitions")
    def test_check_for_partition_false(self, mock_get_partitions):
        mock_get_partitions.return_value = set()
        hook = GlueCatalogHook(region_name="us-east-1")

        assert not hook.check_for_partition("db", "tbl", "expr")

    def test_get_table_exists(self):
        self.client.create_database(DatabaseInput={"Name": DB_NAME})
        self.client.create_table(DatabaseName=DB_NAME, TableInput=TABLE_INPUT)

        result = self.hook.get_table(DB_NAME, TABLE_NAME)

        assert result["Name"] == TABLE_INPUT["Name"]
        assert result["StorageDescriptor"]["Location"] == TABLE_INPUT["StorageDescriptor"]["Location"]

    def test_get_table_not_exists(self):
        self.client.create_database(DatabaseInput={"Name": DB_NAME})
        self.client.create_table(DatabaseName=DB_NAME, TableInput=TABLE_INPUT)

        with pytest.raises(botocore.exceptions.ClientError) as err_ctx:
            self.hook.get_table(DB_NAME, "dummy_table")
        assert err_ctx.value.response["Error"]["Code"] == "EntityNotFoundException"

    def test_get_table_location(self):
        self.client.create_database(DatabaseInput={"Name": DB_NAME})
        self.client.create_table(DatabaseName=DB_NAME, TableInput=TABLE_INPUT)

        result = self.hook.get_table_location(DB_NAME, TABLE_NAME)
        assert result == TABLE_INPUT["StorageDescriptor"]["Location"]

    def test_get_partition(self):
        self.client.create_database(DatabaseInput={"Name": DB_NAME})
        self.client.create_table(DatabaseName=DB_NAME, TableInput=TABLE_INPUT)
        self.client.create_partition(
            DatabaseName=DB_NAME, TableName=TABLE_NAME, PartitionInput=PARTITION_INPUT
        )

        result = self.hook.get_partition(DB_NAME, TABLE_NAME, PARTITION_INPUT["Values"])

        assert result["Values"] == PARTITION_INPUT["Values"]
        assert result["DatabaseName"] == DB_NAME
        assert result["TableName"] == TABLE_INPUT["Name"]

    @mock.patch.object(GlueCatalogHook, "get_conn")
    def test_get_partition_with_client_error(self, mocked_connection):
        mocked_client = mock.Mock()
        mocked_client.get_partition.side_effect = ClientError({}, "get_partition")
        mocked_connection.return_value = mocked_client

        with pytest.raises(AirflowException):
            self.hook.get_partition(DB_NAME, TABLE_NAME, PARTITION_INPUT["Values"])

        mocked_client.get_partition.assert_called_once_with(
            DatabaseName=DB_NAME, TableName=TABLE_NAME, PartitionValues=PARTITION_INPUT["Values"]
        )

    def test_create_partition(self):
        self.client.create_database(DatabaseInput={"Name": DB_NAME})
        self.client.create_table(DatabaseName=DB_NAME, TableInput=TABLE_INPUT)

        result = self.hook.create_partition(DB_NAME, TABLE_NAME, PARTITION_INPUT)

        assert result

    @mock.patch.object(GlueCatalogHook, "get_conn")
    def test_create_partition_with_client_error(self, mocked_connection):
        mocked_client = mock.Mock()
        mocked_client.create_partition.side_effect = ClientError({}, "create_partition")
        mocked_connection.return_value = mocked_client

        with pytest.raises(AirflowException):
            self.hook.create_partition(DB_NAME, TABLE_NAME, PARTITION_INPUT)

        mocked_client.create_partition.assert_called_once_with(
            DatabaseName=DB_NAME, TableName=TABLE_NAME, PartitionInput=PARTITION_INPUT
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import json
import os
from collections.abc import Callable, Sequence
from copy import copy
from datetime import datetime
from decimal import Decimal
from functools import cached_property
from tempfile import NamedTemporaryFile
from typing import IO, TYPE_CHECKING, Any
from uuid import uuid4

from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook
from airflow.providers.amazon.aws.hooks.dynamodb import DynamoDBHook
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.transfers.base import AwsToAwsBaseOperator
from airflow.utils.helpers import prune_dict

if TYPE_CHECKING:
    from airflow.providers.amazon.version_compat import ArgNotSet
    from airflow.sdk import Context


class JSONEncoder(json.JSONEncoder):
    """Custom json encoder implementation."""

    def default(self, obj):
        """Convert decimal objects in a json serializable format."""
        if isinstance(obj, Decimal):
            return float(obj)
        return super().default(obj)


def _convert_item_to_json_bytes(item: dict[str, Any]) -> bytes:
    return (json.dumps(item, cls=JSONEncoder) + "\n").encode("utf-8")


def _upload_file_to_s3(
    file_obj: IO,
    bucket_name: str,
    s3_key_prefix: str,
    aws_conn_id: str | None | ArgNotSet = AwsBaseHook.default_conn_name,
) -> None:
    s3_client = S3Hook(aws_conn_id=aws_conn_id).get_conn()
    file_obj.seek(0)
    s3_client.upload_file(
        Filename=file_obj.name,
        Bucket=bucket_name,
        Key=s3_key_prefix + str(uuid4()),
    )


class DynamoDBToS3Operator(AwsToAwsBaseOperator):
    """
    Replicates records from a DynamoDB table to S3.

    It scans a DynamoDB table and writes the received records to a file
    on the local filesystem. It flushes the file to S3 once the file size
    exceeds the file size limit specified by the user.

    Users can also specify a filtering criteria using dynamodb_scan_kwargs
    to only replicate records that satisfy the criteria.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/transfer:DynamoDBToS3Operator`

    :param dynamodb_table_name: Dynamodb table to replicate data from
    :param s3_bucket_name: S3 bucket to replicate data to
    :param file_size: Flush file to s3 if file size >= file_size
    :param dynamodb_scan_kwargs: kwargs pass to
        <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb.html#DynamoDB.Table.scan>
    :param s3_key_prefix: Prefix of s3 object key
    :param process_func: How we transform a dynamodb item to bytes. By default, we dump the json
    :param point_in_time_export: Boolean value indicating the operator to use 'scan' or 'point in time export'
    :param export_time: Time in the past from which to export table data, counted in seconds from the start of
     the Unix epoch. The table export will be a snapshot of the table's state at this point in time.
    :param export_format: The format for the exported data. Valid values for ExportFormat are DYNAMODB_JSON
     or ION.
    :param export_table_to_point_in_time_kwargs: extra parameters for the boto3
        `export_table_to_point_in_time` function all. e.g. `ExportType`, `IncrementalExportSpecification`
    :param check_interval: The amount of time in seconds to wait between attempts. Only if ``export_time`` is
        provided.
    :param max_attempts: The maximum number of attempts to be made. Only if ``export_time`` is provided.
    """

    template_fields: Sequence[str] = (
        *AwsToAwsBaseOperator.template_fields,
        "dynamodb_table_name",
        "s3_bucket_name",
        "file_size",
        "dynamodb_scan_kwargs",
        "s3_key_prefix",
        "export_time",
        "export_format",
        "export_table_to_point_in_time_kwargs",
        "check_interval",
        "max_attempts",
    )

    template_fields_renderers = {
        "dynamodb_scan_kwargs": "json",
        "export_table_to_point_in_time_kwargs": "json",
    }

    def __init__(
        self,
        *,
        dynamodb_table_name: str,
        s3_bucket_name: str,
        file_size: int = 1000,
        dynamodb_scan_kwargs: dict[str, Any] | None = None,
        s3_key_prefix: str = "",
        process_func: Callable[[dict[str, Any]], bytes] = _convert_item_to_json_bytes,
        point_in_time_export: bool = False,
        export_time: datetime | None = None,
        export_format: str = "DYNAMODB_JSON",
        export_table_to_point_in_time_kwargs: dict | None = None,
        check_interval: int = 30,
        max_attempts: int = 60,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.file_size = file_size
        self.process_func = process_func
        self.dynamodb_table_name = dynamodb_table_name
        self.dynamodb_scan_kwargs = dynamodb_scan_kwargs
        self.s3_bucket_name = s3_bucket_name
        self.s3_key_prefix = s3_key_prefix
        self.point_in_time_export = point_in_time_export
        self.export_time = export_time
        self.export_format = export_format
        self.export_table_to_point_in_time_kwargs = export_table_to_point_in_time_kwargs or {}
        self.check_interval = check_interval
        self.max_attempts = max_attempts

    @cached_property
    def hook(self):
        """Create DynamoDBHook."""
        return DynamoDBHook(aws_conn_id=self.source_aws_conn_id)

    def execute(self, context: Context) -> None:
        # There are 2 separate export to point in time configuration:
        # 1. Full export, which takes the export_time arg.
        # 2. Incremental export, which takes the incremental_export_... args
        # Hence export time could not be used as the proper indicator for the `_export_table_to_point_in_time`
        # function. This change introduces a new boolean, as the indicator for whether the operator scans
        # and export entire data or using the point in time functionality.
        if self.point_in_time_export or self.export_time:
            self._export_table_to_point_in_time()
        else:
            self._export_entire_data()

    def _export_table_to_point_in_time(self):
        """
        Export data to point in time.

        Full export exports data from start of epoc till `export_time`.
        Table export will be a snapshot of the table's state at this point in time.

        Incremental export exports the data from a specific datetime to a specific datetime


        Note: S3BucketOwner is a required parameter when exporting to a S3 bucket in another account.
        """
        if self.export_time and self.export_time > datetime.now(self.export_time.tzinfo):
            raise ValueError("The export_time parameter cannot be a future time.")

        client = self.hook.conn.meta.client
        table_description = client.describe_table(TableName=self.dynamodb_table_name)

        export_table_to_point_in_time_base_args = {
            "TableArn": table_description.get("Table", {}).get("TableArn"),
            "ExportTime": self.export_time,
            "S3Bucket": self.s3_bucket_name,
            "S3Prefix": self.s3_key_prefix,
            "ExportFormat": self.export_format,
        }
        export_table_to_point_in_time_args = {
            **export_table_to_point_in_time_base_args,
            **self.export_table_to_point_in_time_kwargs,
        }

        args_filtered = prune_dict(export_table_to_point_in_time_args)

        response = client.export_table_to_point_in_time(**args_filtered)
        waiter = self.hook.get_waiter("export_table")
        export_arn = response.get("ExportDescription", {}).get("ExportArn")
        waiter.wait(
            ExportArn=export_arn,
            WaiterConfig={"Delay": self.check_interval, "MaxAttempts": self.max_attempts},
        )

    def _export_entire_data(self):
        """Export all data from the table."""
        table = self.hook.conn.Table(self.dynamodb_table_name)
        scan_kwargs = copy(self.dynamodb_scan_kwargs) if self.dynamodb_scan_kwargs else {}
        err = None
        f: IO[Any]
        with NamedTemporaryFile() as f_tmp:
            try:
                f = self._scan_dynamodb_and_upload_to_s3(f_tmp, scan_kwargs, table)
            except Exception as e:
                err = e
                raise e
            finally:
                if err is None:
                    _upload_file_to_s3(f, self.s3_bucket_name, self.s3_key_prefix, self.dest_aws_conn_id)

    def _scan_dynamodb_and_upload_to_s3(self, temp_file: IO, scan_kwargs: dict, table: Any) -> IO:
        while True:
            response = table.scan(**scan_kwargs)
            items = response["Items"]
            for item in items:
                temp_file.write(self.process_func(item))

            if "LastEvaluatedKey" not in response:
                # no more items to scan
                break

            last_evaluated_key = response["LastEvaluatedKey"]
            scan_kwargs["ExclusiveStartKey"] = last_evaluated_key

            # Upload the file to S3 if reach file size limit
            if os.path.getsize(temp_file.name) >= self.file_size:
                _upload_file_to_s3(temp_file, self.s3_bucket_name, self.s3_key_prefix, self.dest_aws_conn_id)
                temp_file.close()

                temp_file = NamedTemporaryFile()
        return temp_file



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock

import pytest

from airflow.providers.amazon.aws.sensors.emr import EmrServerlessApplicationSensor
from airflow.providers.common.compat.sdk import AirflowException


class TestEmrServerlessApplicationSensor:
    def setup_method(self):
        self.app_id = "vzwemreks"
        self.job_run_id = "job1234"
        self.sensor = EmrServerlessApplicationSensor(
            task_id="test_emrcontainer_sensor",
            application_id=self.app_id,
            aws_conn_id="aws_default",
        )

    def set_get_application_return_value(self, return_value: dict[str, str]):
        self.mock_hook = MagicMock()
        self.mock_hook.conn.get_application.return_value = return_value
        self.sensor.hook = self.mock_hook

    def assert_get_application_was_called_once_with_app_id(self):
        self.mock_hook.conn.get_application.assert_called_once_with(applicationId=self.app_id)


class TestPokeReturnValue(TestEmrServerlessApplicationSensor):
    @pytest.mark.parametrize(
        ("state", "expected_result"),
        [
            ("CREATING", False),
            ("STARTING", False),
            ("STOPPING", False),
            ("CREATED", True),
            ("STARTED", True),
        ],
    )
    def test_poke_returns_expected_result_for_states(self, state, expected_result):
        get_application_return_value = {"application": {"state": state}}
        self.set_get_application_return_value(get_application_return_value)
        assert self.sensor.poke(None) == expected_result
        self.assert_get_application_was_called_once_with_app_id()


class TestPokeRaisesAirflowException(TestEmrServerlessApplicationSensor):
    @pytest.mark.parametrize("state", ["STOPPED", "TERMINATED"])
    def test_poke_raises_airflow_exception_with_failure_states(self, state):
        state_details = f"mock {state}"
        exception_msg = f"EMR Serverless application failed: {state_details}"
        get_job_run_return_value = {"application": {"state": state, "stateDetails": state_details}}
        self.set_get_application_return_value(get_job_run_return_value)

        with pytest.raises(AirflowException) as ctx:
            self.sensor.poke(None)

        assert exception_msg == str(ctx.value)
        self.assert_get_application_was_called_once_with_app_id()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.amazon.aws.links.batch import (
    BatchJobDefinitionLink,
    BatchJobDetailsLink,
    BatchJobQueueLink,
)

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS
from unit.amazon.aws.links.test_base_aws import BaseAwsLinksTestCase

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk.execution_time.comms import XComResult

pytestmark = pytest.mark.db_test


class TestBatchJobDefinitionLink(BaseAwsLinksTestCase):
    link_class = BatchJobDefinitionLink

    def test_extra_link(self, mock_supervisor_comms):
        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key=self.link_class.key,
                value={
                    "region_name": "eu-west-1",
                    "aws_domain": self.link_class.get_aws_domain("aws"),
                    "aws_partition": "aws",
                    "job_definition_arn": "arn:fake:jd",
                },
            )
        self.assert_extra_link_url(
            expected_url=(
                "https://console.aws.amazon.com/batch/home?region=eu-west-1#job-definition/detail/arn:fake:jd"
            ),
            region_name="eu-west-1",
            aws_partition="aws",
            job_definition_arn="arn:fake:jd",
        )


class TestBatchJobDetailsLink(BaseAwsLinksTestCase):
    link_class = BatchJobDetailsLink

    def test_extra_link(self, mock_supervisor_comms):
        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key=self.link_class.key,
                value={
                    "region_name": "cn-north-1",
                    "aws_domain": self.link_class.get_aws_domain("aws-cn"),
                    "aws_partition": "aws-cn",
                    "job_id": "fake-id",
                },
            )
        self.assert_extra_link_url(
            expected_url="https://console.amazonaws.cn/batch/home?region=cn-north-1#jobs/detail/fake-id",
            region_name="cn-north-1",
            aws_partition="aws-cn",
            job_id="fake-id",
        )


class TestBatchJobQueueLink(BaseAwsLinksTestCase):
    link_class = BatchJobQueueLink

    def test_extra_link(self, mock_supervisor_comms):
        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.send.return_value = XComResult(
                key=self.link_class.key,
                value={
                    "region_name": "us-east-1",
                    "aws_domain": self.link_class.get_aws_domain("aws"),
                    "aws_partition": "aws",
                    "job_queue_arn": "arn:fake:jq",
                },
            )
        self.assert_extra_link_url(
            expected_url=(
                "https://console.aws.amazon.com/batch/home?region=us-east-1#queues/detail/arn:fake:jq"
            ),
            region_name="us-east-1",
            aws_partition="aws",
            job_queue_arn="arn:fake:jq",
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import AsyncMock

import pytest
from botocore.exceptions import WaiterError

from airflow.providers.amazon.aws.triggers.neptune import (
    NeptuneClusterAvailableTrigger,
    NeptuneClusterInstancesAvailableTrigger,
    NeptuneClusterStoppedTrigger,
)
from airflow.providers.common.compat.sdk import AirflowException
from airflow.triggers.base import TriggerEvent

CLUSTER_ID = "test-cluster"


class TestNeptuneClusterAvailableTrigger:
    def test_serialization(self):
        """
        Asserts that the TaskStateTrigger correctly serializes its arguments
        and classpath.
        """
        trigger = NeptuneClusterAvailableTrigger(db_cluster_id=CLUSTER_ID)
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.neptune.NeptuneClusterAvailableTrigger"
        assert "db_cluster_id" in kwargs
        assert kwargs["db_cluster_id"] == CLUSTER_ID

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.amazon.aws.hooks.neptune.NeptuneHook.get_waiter")
    @mock.patch("airflow.providers.amazon.aws.hooks.neptune.NeptuneHook.get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.return_value.__aenter__.return_value = "available"
        mock_get_waiter().wait = AsyncMock()
        trigger = NeptuneClusterAvailableTrigger(db_cluster_id=CLUSTER_ID)
        generator = trigger.run()
        resp = await generator.asend(None)

        assert resp == TriggerEvent({"status": "success", "db_cluster_id": CLUSTER_ID})
        assert mock_get_waiter().wait.call_count == 1


class TestNeptuneClusterStoppedTrigger:
    def test_serialization(self):
        """
        Asserts that the TaskStateTrigger correctly serializes its arguments
        and classpath.
        """
        trigger = NeptuneClusterStoppedTrigger(db_cluster_id=CLUSTER_ID)
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.neptune.NeptuneClusterStoppedTrigger"
        assert "db_cluster_id" in kwargs
        assert kwargs["db_cluster_id"] == CLUSTER_ID

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.amazon.aws.hooks.neptune.NeptuneHook.get_waiter")
    @mock.patch("airflow.providers.amazon.aws.hooks.neptune.NeptuneHook.get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.return_value.__aenter__.return_value = "stopped"
        mock_get_waiter().wait = AsyncMock()
        trigger = NeptuneClusterStoppedTrigger(db_cluster_id=CLUSTER_ID)
        generator = trigger.run()
        resp = await generator.asend(None)

        assert resp == TriggerEvent({"status": "success", "db_cluster_id": CLUSTER_ID})
        assert mock_get_waiter().wait.call_count == 1


class TestNeptuneClusterInstancesAvailableTrigger:
    def test_serialization(self):
        """
        Asserts that the TaskStateTrigger correctly serializes its arguments
        and classpath.
        """
        trigger = NeptuneClusterInstancesAvailableTrigger(db_cluster_id=CLUSTER_ID)
        classpath, kwargs = trigger.serialize()
        assert (
            classpath
            == "airflow.providers.amazon.aws.triggers.neptune.NeptuneClusterInstancesAvailableTrigger"
        )
        assert "db_cluster_id" in kwargs
        assert kwargs["db_cluster_id"] == CLUSTER_ID

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.amazon.aws.hooks.neptune.NeptuneHook.get_waiter")
    @mock.patch("airflow.providers.amazon.aws.hooks.neptune.NeptuneHook.get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.return_value.__aenter__.return_value = "available"
        mock_get_waiter().wait = AsyncMock()
        trigger = NeptuneClusterInstancesAvailableTrigger(db_cluster_id=CLUSTER_ID)
        generator = trigger.run()
        resp = await generator.asend(None)

        assert resp == TriggerEvent({"status": "success", "db_cluster_id": CLUSTER_ID})
        assert mock_get_waiter().wait.call_count == 1

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.amazon.aws.hooks.neptune.NeptuneHook.get_async_conn")
    async def test_run_fail(self, mock_async_conn):
        a_mock = mock.MagicMock()
        mock_async_conn.return_value.__aenter__.return_value = a_mock
        wait_mock = AsyncMock()
        wait_mock.side_effect = WaiterError("name", "reason", {"test": [{"lastStatus": "my_status"}]})
        a_mock.get_waiter().wait = wait_mock
        trigger = NeptuneClusterInstancesAvailableTrigger(
            db_cluster_id=CLUSTER_ID, waiter_delay=1, waiter_max_attempts=2
        )

        with pytest.raises(AirflowException):
            await trigger.run().asend(None)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime as dt
import json
import logging
import os
from unittest import mock
from unittest.mock import patch

import pytest
import yaml
from botocore.exceptions import ClientError, NoCredentialsError
from semver import VersionInfo

from airflow.configuration import conf
from airflow.executors.base_executor import BaseExecutor
from airflow.models import TaskInstance
from airflow.models.taskinstancekey import TaskInstanceKey
from airflow.providers.amazon.aws.executors.batch import batch_executor, batch_executor_config
from airflow.providers.amazon.aws.executors.batch.batch_executor import (
    AwsBatchExecutor,
    BatchJob,
    BatchJobCollection,
)
from airflow.providers.amazon.aws.executors.batch.utils import (
    CONFIG_DEFAULTS,
    CONFIG_GROUP_NAME,
    AllBatchConfigKeys,
)
from airflow.providers.common.compat.sdk import AirflowException
from airflow.utils.helpers import convert_camel_to_snake
from airflow.utils.state import State
from airflow.version import version as airflow_version_str

from tests_common import RUNNING_TESTS_AGAINST_AIRFLOW_PACKAGES
from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS, AIRFLOW_V_3_1_PLUS

airflow_version = VersionInfo(*map(int, airflow_version_str.split(".")[:3]))
ARN1 = "arn1"

MOCK_JOB_ID = "batch-job-id"


@pytest.fixture
def set_env_vars():
    overrides: dict[tuple[str, str], str] = {
        (CONFIG_GROUP_NAME, AllBatchConfigKeys.REGION_NAME): "us-east-1",
        (CONFIG_GROUP_NAME, AllBatchConfigKeys.JOB_NAME): "some-job-name",
        (CONFIG_GROUP_NAME, AllBatchConfigKeys.JOB_QUEUE): "some-job-queue",
        (CONFIG_GROUP_NAME, AllBatchConfigKeys.JOB_DEFINITION): "some-job-def",
        (CONFIG_GROUP_NAME, AllBatchConfigKeys.MAX_SUBMIT_JOB_ATTEMPTS): "3",
        (CONFIG_GROUP_NAME, AllBatchConfigKeys.CHECK_HEALTH_ON_STARTUP): "True",
    }
    with conf_vars(overrides):
        yield


@pytest.fixture
def mock_executor(set_env_vars) -> AwsBatchExecutor:
    """Mock Batch Executor to a repeatable starting state."""
    executor = AwsBatchExecutor()
    executor.IS_BOTO_CONNECTION_HEALTHY = True

    # Replace boto3 Batch client with mock.
    batch_mock = mock.Mock(spec=executor.batch)
    submit_job_ret_val = {"tasks": [{"taskArn": ARN1}], "failures": []}
    batch_mock.submit_job.return_value = submit_job_ret_val
    executor.batch = batch_mock

    return executor


@pytest.fixture(autouse=True)
def mock_airflow_key():
    return mock.Mock(spec=list)


@pytest.fixture(autouse=True)
def mock_cmd():
    return mock.Mock(spec=list)


class TestBatchJobCollection:
    """Tests BatchJobCollection Class"""

    @pytest.fixture(autouse=True)
    def _setup_method(self):
        """
        Create a BatchJobCollection object and add 2 airflow tasks. Populates self.collection,
        self.first/second_task, self.first/second_airflow_key, and self.first/second_airflow_cmd.
        """
        self.collection = BatchJobCollection()
        # Add first task
        self.first_job_id = "001"
        self.first_airflow_key = mock.Mock(spec=tuple)
        self.collection.add_job(
            job_id=self.first_job_id,
            airflow_task_key=self.first_airflow_key,
            airflow_cmd="command1",
            queue="queue1",
            exec_config={},
            attempt_number=1,
        )
        # Add second task
        self.second_job_id = "002"
        self.second_airflow_key = mock.Mock(spec=tuple)
        self.collection.add_job(
            job_id=self.second_job_id,
            airflow_task_key=self.second_airflow_key,
            airflow_cmd="command2",
            queue="queue2",
            exec_config={},
            attempt_number=1,
        )

    def test_get_and_add(self):
        """Test add_task, task_by_arn, cmd_by_key"""
        assert len(self.collection) == 2

    def test_list(self):
        """Test get_all_arns() and get_all_task_keys()"""
        # Check basic list by ARNs & airflow-task-keys
        assert self.collection.get_all_jobs() == [self.first_job_id, self.second_job_id]

    def test_pop(self):
        """Test pop_by_key()"""
        # pop first task & ensure that it's removed
        self.collection.pop_by_id(self.first_job_id)
        assert len(self.collection) == 1
        assert self.collection.get_all_jobs() == [self.second_job_id]


class TestBatchJob:
    """Tests the BatchJob DTO"""

    @pytest.fixture(autouse=True)
    def _setup_method(self):
        self.all_statuses = ["SUBMITTED", "PENDING", "RUNNABLE", "STARTING", "RUNNING", "SUCCEEDED", "FAILED"]
        self.running = "RUNNING"
        self.success = "SUCCEEDED"
        self.failed = "FAILED"

    def test_queued_jobs(self):
        """Jobs that are pending launch identified as 'queued'"""
        for status in self.all_statuses:
            if status not in (self.success, self.failed, self.running):
                job = BatchJob("id", status)
                job_state = job.get_job_state()
                assert job_state not in (State.RUNNING, State.FAILED, State.SUCCESS)
                assert job_state == State.QUEUED

    def test_running_jobs(self):
        """Jobs that have been launched are identified as 'running'"""
        assert self.running in self.all_statuses
        running_job = BatchJob("AAA", self.running)
        assert running_job.get_job_state() == State.RUNNING

    def test_success_jobs(self):
        """Jobs that have been launched are identified as 'SUCCEEDED'"""
        assert self.success in self.all_statuses

        success_job = BatchJob("BBB", self.success)
        assert success_job.get_job_state() == State.SUCCESS

    def test_failed_jobs(self):
        """Jobs that have been launched are identified as 'FAILED'"""
        assert self.failed in self.all_statuses
        running_job = BatchJob("CCC", self.failed)
        assert running_job.get_job_state() == State.FAILED


class TestAwsBatchExecutor:
    """Tests the AWS Batch Executor itself"""

    def test_execute(self, mock_executor):
        """Test execution from end-to-end"""
        airflow_key = mock.Mock(spec=tuple)
        airflow_cmd = ["1", "2"]

        mock_executor.batch.submit_job.return_value = {"jobId": MOCK_JOB_ID, "jobName": "some-job-name"}

        mock_executor.execute_async(airflow_key, airflow_cmd)
        assert len(mock_executor.pending_jobs) == 1
        mock_executor.attempt_submit_jobs()
        mock_executor.batch.submit_job.assert_called_once()
        assert len(mock_executor.active_workers) == 1

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow 3+")
    @mock.patch("airflow.providers.amazon.aws.executors.batch.batch_executor.AwsBatchExecutor.running_state")
    def test_task_sdk(self, running_state_mock, mock_airflow_key, mock_executor, mock_cmd):
        """Test task sdk execution from end-to-end."""
        from airflow.executors.workloads import ExecuteTask

        workload = mock.Mock(spec=ExecuteTask)
        workload.ti = mock.Mock(spec=TaskInstance)
        workload.ti.key = mock_airflow_key()
        tags_exec_config = [{"key": "FOO", "value": "BAR"}]
        workload.ti.executor_config = {"tags": tags_exec_config}
        ser_workload = json.dumps({"test_key": "test_value"})
        workload.model_dump_json.return_value = ser_workload

        mock_executor.queue_workload(workload, mock.Mock())

        mock_executor.batch.submit_job.return_value = {"jobId": ARN1, "jobName": "some-job-name"}

        assert mock_executor.queued_tasks[workload.ti.key] == workload
        assert len(mock_executor.pending_jobs) == 0
        assert len(mock_executor.running) == 0
        mock_executor._process_workloads([workload])
        assert len(mock_executor.queued_tasks) == 0
        assert len(mock_executor.running) == 1
        assert workload.ti.key in mock_executor.running
        assert len(mock_executor.pending_jobs) == 1
        assert mock_executor.pending_jobs[0].command == [
            "python",
            "-m",
            "airflow.sdk.execution_time.execute_workload",
            "--json-string",
            '{"test_key": "test_value"}',
        ]

        mock_executor.attempt_submit_jobs()
        mock_executor.batch.submit_job.assert_called_once()
        assert len(mock_executor.pending_jobs) == 0
        mock_executor.batch.submit_job.assert_called_once_with(
            jobDefinition="some-job-def",
            jobName="some-job-name",
            jobQueue="some-job-queue",
            tags=tags_exec_config,
            containerOverrides={
                "command": [
                    "python",
                    "-m",
                    "airflow.sdk.execution_time.execute_workload",
                    "--json-string",
                    ser_workload,
                ],
                "environment": [
                    {
                        "name": "AIRFLOW_IS_EXECUTOR_CONTAINER",
                        "value": "true",
                    },
                ],
            },
        )

        # Task is stored in active worker.
        assert len(mock_executor.active_workers) == 1
        # Get the job_id for this task key
        job_id = next(
            job_id for job_id, key in mock_executor.active_workers.id_to_key.items() if key == workload.ti.key
        )
        assert job_id == ARN1
        running_state_mock.assert_called_once_with(workload.ti.key, ARN1)

    @mock.patch.object(batch_executor, "calculate_next_attempt_delay", return_value=dt.timedelta(seconds=0))
    def test_attempt_all_jobs_when_some_jobs_fail(self, _, mock_executor):
        """
        Test how jobs are tried when one job fails, but others pass.

        The expected behaviour is that in one sync() iteration, all the jobs are attempted
        exactly once. Successful jobs are removed from pending_jobs to active_workers, and
        failed jobs are added back to the pending_jobs queue to be run in the next iteration.
        """
        airflow_key = TaskInstanceKey("a", "b", "c", 1, -1)
        airflow_cmd1 = ["1", "2"]
        airflow_cmd2 = ["3", "4"]
        airflow_commands = [airflow_cmd1, airflow_cmd2]
        responses = [Exception("Failure 1"), {"jobId": "job-2"}]

        submit_job_args = {
            "jobDefinition": "some-job-def",
            "jobName": "some-job-name",
            "jobQueue": "some-job-queue",
            "containerOverrides": {
                "command": ["command"],
                "environment": [{"name": "AIRFLOW_IS_EXECUTOR_CONTAINER", "value": "true"}],
            },
        }
        mock_executor.execute_async(airflow_key, airflow_cmd1)
        mock_executor.execute_async(airflow_key, airflow_cmd2)
        assert len(mock_executor.pending_jobs) == 2

        mock_executor.batch.submit_job.side_effect = responses
        mock_executor.attempt_submit_jobs()

        for i in range(2):
            submit_job_args["containerOverrides"]["command"] = airflow_commands[i]
            assert mock_executor.batch.submit_job.call_args_list[i].kwargs == submit_job_args
        assert len(mock_executor.pending_jobs) == 1
        mock_executor.pending_jobs[0].command == airflow_cmd1
        assert len(mock_executor.active_workers.get_all_jobs()) == 1

        # Add more tasks to pending_jobs. This simulates tasks being scheduled by Airflow
        airflow_cmd3 = ["5", "6"]
        airflow_cmd4 = ["7", "8"]
        airflow_commands.extend([airflow_cmd1, airflow_cmd3, airflow_cmd4])
        responses.extend([Exception("Failure 1"), {"jobId": "job-3"}, {"jobId": "job-4"}])
        mock_executor.execute_async(airflow_key, airflow_cmd3)
        mock_executor.execute_async(airflow_key, airflow_cmd4)
        assert len(mock_executor.pending_jobs) == 3

        mock_executor.attempt_submit_jobs()
        assert len(mock_executor.pending_jobs) == 1
        assert len(mock_executor.active_workers.get_all_jobs()) == 3

        for i in range(2, 5):
            submit_job_args["containerOverrides"]["command"] = airflow_commands[i]
            assert mock_executor.batch.submit_job.call_args_list[i].kwargs == submit_job_args
        assert len(mock_executor.pending_jobs) == 1
        mock_executor.pending_jobs[0].command == airflow_cmd1
        assert len(mock_executor.active_workers.get_all_jobs()) == 3

        airflow_commands.append(airflow_cmd1)
        responses.append(Exception("Failure 1"))

        mock_executor.attempt_submit_jobs()
        submit_job_args["containerOverrides"]["command"] = airflow_commands[0]
        assert mock_executor.batch.submit_job.call_args_list[5].kwargs == submit_job_args
        if airflow_version >= (2, 10, 0):
            log_record = mock_executor._task_event_logs[0]
            assert log_record.event == "batch job submit failure"

    @mock.patch.object(batch_executor, "calculate_next_attempt_delay", return_value=dt.timedelta(seconds=0))
    def test_attempt_all_jobs_when_jobs_fail(self, _, mock_executor):
        """
        Test job retry behaviour when jobs fail validation.

        Test that when a job fails with a client sided exception, all the jobs are
        attempted once. If all jobs fail, then the length of pending tasks should not change,
        until all the tasks have been attempted the maximum number of times.
        """
        airflow_key = TaskInstanceKey("a", "b", "c", 1, -1)
        airflow_cmd1 = ["1", "2"]
        airflow_cmd2 = ["3", "4"]
        commands = [airflow_cmd1, airflow_cmd2]
        failures = [Exception("Failure 1"), Exception("Failure 2")]
        submit_job_args = {
            "containerOverrides": {
                "command": ["command"],
                "environment": [{"name": "AIRFLOW_IS_EXECUTOR_CONTAINER", "value": "true"}],
            },
            "jobDefinition": "some-job-def",
            "jobName": "some-job-name",
            "jobQueue": "some-job-queue",
        }
        mock_executor.execute_async(airflow_key, airflow_cmd1)
        mock_executor.execute_async(airflow_key, airflow_cmd2)
        assert len(mock_executor.pending_jobs) == 2

        mock_executor.batch.submit_job.side_effect = failures
        mock_executor.attempt_submit_jobs()

        for i in range(2):
            submit_job_args["containerOverrides"]["command"] = commands[i]
            assert mock_executor.batch.submit_job.call_args_list[i].kwargs == submit_job_args
        assert len(mock_executor.pending_jobs) == 2

        mock_executor.batch.submit_job.side_effect = failures
        mock_executor.attempt_submit_jobs()
        for i in range(2):
            submit_job_args["containerOverrides"]["command"] = commands[i]
            assert mock_executor.batch.submit_job.call_args_list[i].kwargs == submit_job_args
        assert len(mock_executor.pending_jobs) == 2

        mock_executor.batch.submit_job.side_effect = failures
        mock_executor.attempt_submit_jobs()
        if VersionInfo.parse(str(airflow_version)) >= (2, 10, 0):
            events = [(x.event, x.task_id, x.try_number) for x in mock_executor._task_event_logs]
            assert events == [("batch job submit failure", "b", 1)] * 2

    def test_attempt_submit_jobs_failure(self, mock_executor):
        mock_executor.batch.submit_job.side_effect = NoCredentialsError()
        mock_executor.execute_async("airflow_key", "airflow_cmd")
        assert len(mock_executor.pending_jobs) == 1
        with pytest.raises(NoCredentialsError, match="Unable to locate credentials"):
            mock_executor.attempt_submit_jobs()
        assert len(mock_executor.active_workers.get_all_jobs()) == 0
        assert len(mock_executor.pending_jobs) == 1
        mock_resp = {
            "Error": {
                "Code": "ExpiredTokenException",
                "Message": "Expired token error",
            },
        }
        mock_executor.batch.submit_job.side_effect = ClientError(mock_resp, "test_submit_jobs")
        with pytest.raises(ClientError, match="Expired token error"):
            mock_executor.attempt_submit_jobs()
        assert len(mock_executor.active_workers.get_all_jobs()) == 0
        assert len(mock_executor.pending_jobs) == 1

    @mock.patch.object(batch_executor, "calculate_next_attempt_delay", return_value=dt.timedelta(seconds=0))
    def test_task_retry_on_api_failure(self, _, mock_executor, caplog):
        """Test API failure retries"""
        airflow_keys = ["TaskInstanceKey1", "TaskInstanceKey2"]
        airflow_cmds = [["1", "2"], ["3", "4"]]

        mock_executor.execute_async(airflow_keys[0], airflow_cmds[0])
        mock_executor.execute_async(airflow_keys[1], airflow_cmds[1])
        assert len(mock_executor.pending_jobs) == 2
        jobs = [
            {
                "jobName": "job-1",
                "jobId": "job-1",
                "status": "FAILED",
                "statusReason": "Test Failure for job1",
            },
            {
                "jobName": "job-2",
                "jobId": "job-2",
                "status": "FAILED",
                "statusReason": "Test Failure for job2",
            },
        ]
        mock_executor.batch.describe_jobs.return_value = {"jobs": jobs}
        mock_executor.batch.submit_job.side_effect = [
            {"jobId": "job-1"},
            {"jobId": "job-2"},
            {"jobId": "job-1"},
            {"jobId": "job-2"},
            {"jobId": "job-1"},
            {"jobId": "job-2"},
        ]
        mock_executor.attempt_submit_jobs()

        assert len(mock_executor.active_workers.get_all_jobs()) == 2
        assert len(mock_executor.pending_jobs) == 0

        mock_executor.sync_running_jobs()
        for i in range(2):
            assert (
                f"Airflow task {airflow_keys[i]} failed due to {jobs[i]['statusReason']}. Failure 1 out of {mock_executor.max_submit_job_attempts} occurred on {jobs[i]['jobId']}. Rescheduling."
                in caplog.messages[i]
            )

        caplog.clear()
        mock_executor.attempt_submit_jobs()
        mock_executor.sync_running_jobs()
        for i in range(2):
            assert (
                f"Airflow task {airflow_keys[i]} failed due to {jobs[i]['statusReason']}. Failure 2 out of {mock_executor.max_submit_job_attempts} occurred on {jobs[i]['jobId']}. Rescheduling."
                in caplog.messages[i]
            )

        caplog.clear()
        mock_executor.attempt_submit_jobs()
        mock_executor.sync_running_jobs()
        for i in range(2):
            assert (
                f"Airflow task {airflow_keys[i]} has failed a maximum of {mock_executor.max_submit_job_attempts} times. Marking as failed"
                in caplog.text
            )

    @mock.patch("airflow.providers.amazon.aws.executors.batch.batch_executor.exponential_backoff_retry")
    def test_sync_unhealthy_boto_connection(self, mock_exponentional_backoff_retry, mock_executor):
        mock_exponentional_backoff_retry.return_value = None
        mock_executor.IS_BOTO_CONNECTION_HEALTHY = False
        mock_executor.sync()
        mock_exponentional_backoff_retry.assert_called_once()
        assert mock_executor.IS_BOTO_CONNECTION_HEALTHY is False

    def test_sync_running_jobs_no_jobs(self, mock_executor, caplog):
        caplog.set_level("DEBUG")
        assert len(mock_executor.active_workers.get_all_jobs()) == 0
        mock_executor.sync_running_jobs()
        assert "No active Airflow tasks, skipping sync" in caplog.messages[0]

    def test_sync_client_error(self, mock_executor, caplog):
        mock_executor.execute_async("airflow_key", "airflow_cmd")
        assert len(mock_executor.pending_jobs) == 1
        mock_resp = {
            "Error": {
                "Code": "ExpiredTokenException",
                "Message": "Expired token error",
            },
        }
        caplog.set_level("WARNING")
        mock_executor.batch.submit_job.side_effect = ClientError(mock_resp, "test-sync")
        mock_executor.sync()
        assert "AWS credentials are either missing or expired" in caplog.messages[0]

    def test_sync_exception(self, mock_executor, caplog):
        mock_executor.active_workers.add_job(
            job_id="job_id",
            airflow_task_key="airflow_key",
            airflow_cmd="command",
            queue="queue",
            exec_config={},
            attempt_number=1,
        )
        assert len(mock_executor.active_workers.get_all_jobs()) == 1
        caplog.set_level("ERROR")
        mock_executor.batch.describe_jobs.side_effect = Exception("test-exception")
        mock_executor.sync()
        assert f"Failed to sync {mock_executor.__class__.__name__}" in caplog.messages[0]

    @mock.patch.object(BaseExecutor, "fail")
    @mock.patch.object(BaseExecutor, "success")
    def test_sync(self, success_mock, fail_mock, mock_airflow_key, mock_executor):
        """Test sync from end-to-end. Mocks a successful job & makes sure it's removed"""
        self._mock_sync(executor=mock_executor, airflow_key=mock_airflow_key())
        mock_executor.sync()

        mock_executor.batch.describe_jobs.assert_called_once()

        # task is not stored in active workers
        assert len(mock_executor.active_workers) == 0
        # Task is immediately succeeded
        success_mock.assert_called_once()
        assert fail_mock.call_count == 0

    @mock.patch.object(BaseExecutor, "fail")
    @mock.patch.object(BaseExecutor, "success")
    @mock.patch.object(batch_executor, "calculate_next_attempt_delay", return_value=dt.timedelta(seconds=0))
    def test_failed_sync(self, _, success_mock, fail_mock, mock_airflow_key, mock_executor):
        """Test failure states"""
        self._mock_sync(
            executor=mock_executor,
            airflow_key=mock_airflow_key(),
            status="FAILED",
            attempt_number=2,
        )

        mock_executor.sync()

        mock_executor.batch.describe_jobs.assert_called_once()

        # task is not stored in active workers
        assert len(mock_executor.active_workers) == 0
        # Task is immediately failed
        fail_mock.assert_called_once()
        assert success_mock.call_count == 0

    def test_start_failure_with_invalid_permissions(self, set_env_vars):
        executor = AwsBatchExecutor()

        # Replace boto3 Batch client with mock.
        batch_mock = mock.Mock(spec=executor.batch)
        batch_mock.describe_jobs.return_value = {}
        executor.batch = batch_mock

        mock_resp = {
            "Error": {
                "Code": "AccessDeniedException",
                "Message": "is not authorized to perform: batch:DescribeJobs on resource:",
            }
        }
        batch_mock.describe_jobs.side_effect = ClientError(mock_resp, "DescribeJobs")

        executor.batch = batch_mock

        with pytest.raises(AirflowException, match=mock_resp["Error"]["Message"]):
            executor.start()

    def test_start_success(self, set_env_vars, caplog):
        executor = AwsBatchExecutor()

        # Replace boto3 Batch client with mock.
        batch_mock = mock.Mock(spec=executor.batch)
        batch_mock.describe_jobs.return_value = {}
        executor.batch = batch_mock

        caplog.clear()
        with caplog.at_level(logging.DEBUG):
            executor.start()

        assert "succeeded" in caplog.text

    def test_health_check_failure(self, mock_executor, set_env_vars):
        mock_executor.batch.describe_jobs.side_effect = Exception("Test_failure")
        executor = AwsBatchExecutor()
        batch_mock = mock.Mock(spec=executor.batch)
        batch_mock.describe_jobs.side_effect = Exception("Test_failure")
        executor.batch = batch_mock

        with pytest.raises(Exception, match="Test_failure"):
            executor.start()

    def test_start_health_check_config(self, set_env_vars):
        executor = AwsBatchExecutor()

        # Replace boto3 Batch client with mock.
        batch_mock = mock.Mock(spec=executor.batch)
        batch_mock.describe_jobs.side_effect = {}
        executor.batch = batch_mock

        env_var_key = f"AIRFLOW__{CONFIG_GROUP_NAME}__{AllBatchConfigKeys.CHECK_HEALTH_ON_STARTUP}".upper()
        with mock.patch.dict(os.environ, {env_var_key: "False"}):
            executor.start()

        batch_mock.describe_jobs.assert_not_called()

    def test_terminate(self, mock_airflow_key, mock_executor):
        """Test that executor can shut everything down; forcing all tasks to unnaturally exit"""
        self._mock_sync(executor=mock_executor, airflow_key=mock_airflow_key(), status="FAILED")

        mock_executor.terminate()
        mock_executor.batch.terminate_job.assert_called_once()

    def test_terminate_failure(self, mock_executor, caplog):
        mock_executor.active_workers.add_job(
            job_id="job_id",
            airflow_task_key="airflow_key",
            airflow_cmd="command",
            queue="queue",
            exec_config={},
            attempt_number=1,
        )
        assert len(mock_executor.active_workers.get_all_jobs()) == 1
        caplog.set_level("ERROR")
        mock_executor.batch.terminate_job.side_effect = Exception("test-exception")
        mock_executor.terminate()
        assert f"Failed to terminate {mock_executor.__class__.__name__}" in caplog.messages[0]

    def test_end(self, mock_airflow_key, mock_executor):
        """The end() function should call sync 3 times, and the task should fail on the 3rd call"""
        self.sync_call_count = 0
        sync_func = mock_executor.sync
        self._mock_sync(executor=mock_executor, airflow_key=mock_airflow_key(), status="RUNNING")

        def sync_mock():
            """This is to count the number of times sync is called. On the 3rd time, mock the job to fail"""
            if self.sync_call_count >= 2:
                self._mock_sync(executor=mock_executor, airflow_key=mock_airflow_key(), status="FAILED")
            sync_func()
            self.sync_call_count += 1

        mock_executor.sync = sync_mock
        self._mock_sync(executor=mock_executor, airflow_key=mock_airflow_key(), status="RUNNING")
        mock_executor.end(heartbeat_interval=0)
        assert self.sync_call_count == 3
        mock_executor.sync = sync_func

    def _mock_sync(
        self,
        executor: AwsBatchExecutor,
        airflow_key: str,
        job_id: str = MOCK_JOB_ID,
        status: str = "SUCCEEDED",
        status_reason: str = "",
        attempt_number: int = 0,
    ):
        """
        This function is not mocking sync so much as it is preparing for
        sync to be called. It adds a job to active_workers and mocks the describe_jobs API call.
        """
        executor.active_workers.add_job(
            job_id=job_id,
            airflow_task_key=airflow_key,
            airflow_cmd="airflow_cmd",
            queue="queue",
            exec_config={},
            attempt_number=attempt_number,
        )

        after_batch_job = {
            "jobName": "some-job-queue",
            "jobId": job_id,
            "jobQueue": "some-job-queue",
            "status": status,
            "statusReason": status_reason,
            "createdAt": dt.datetime.now().timestamp(),
            "jobDefinition": "some-job-def",
        }
        executor.batch.describe_jobs.return_value = {"jobs": [after_batch_job]}

    @pytest.mark.skip(reason="Adopting task instances hasn't been ported over to Airflow 3 yet")
    def test_try_adopt_task_instances(self, mock_executor):
        """Test that executor can adopt orphaned task instances from a SchedulerJob shutdown event."""
        mock_executor.batch.describe_jobs.return_value = {
            "jobs": [
                {"jobId": "001", "status": "SUCCEEDED"},
                {"jobId": "002", "status": "SUCCEEDED"},
            ],
        }

        orphaned_tasks = [
            mock.Mock(spec=TaskInstance),
            mock.Mock(spec=TaskInstance),
            mock.Mock(spec=TaskInstance),
        ]
        orphaned_tasks[0].external_executor_id = "001"  # Matches a running task_arn
        orphaned_tasks[1].external_executor_id = "002"  # Matches a running task_arn
        orphaned_tasks[2].external_executor_id = None  # One orphaned task has no external_executor_id
        for task in orphaned_tasks:
            task.try_number = 1

        not_adopted_tasks = mock_executor.try_adopt_task_instances(orphaned_tasks)

        mock_executor.batch.describe_jobs.assert_called_once()
        # Two of the three tasks should be adopted.
        assert len(orphaned_tasks) - 1 == len(mock_executor.active_workers)
        # The remaining one task is unable to be adopted.
        assert len(not_adopted_tasks) == 1

    @pytest.mark.skipif(not AIRFLOW_V_3_1_PLUS, reason="Multi-team support requires Airflow 3.1+")
    def test_team_config(self):
        """Test that the executor uses team-specific configuration when provided via self.conf."""
        # Team name to be used throughout
        team_name = "team_a"
        # Patch environment to include two sets of configs for the Batch executor. One that is related to a
        # team and one that is not. Then we will create two executors (one with a team and one without) and
        # ensure the correct configs are used.
        config_overrides = [
            (f"AIRFLOW__{CONFIG_GROUP_NAME}__{AllBatchConfigKeys.JOB_QUEUE}", "some-job-queue"),
            (f"AIRFLOW__{CONFIG_GROUP_NAME}__{AllBatchConfigKeys.JOB_DEFINITION}", "some-job-def"),
            (f"AIRFLOW__{CONFIG_GROUP_NAME}__{AllBatchConfigKeys.JOB_NAME}", "some-job-name"),
            (f"AIRFLOW__{CONFIG_GROUP_NAME}__{AllBatchConfigKeys.REGION_NAME}", "us-west-1"),
            # team Config
            (
                f"AIRFLOW__{team_name}___{CONFIG_GROUP_NAME}__{AllBatchConfigKeys.JOB_QUEUE}",
                "team_a_job_queue",
            ),
            (
                f"AIRFLOW__{team_name}___{CONFIG_GROUP_NAME}__{AllBatchConfigKeys.JOB_DEFINITION}",
                "team_a_job_def",
            ),
            (f"AIRFLOW__{team_name}___{CONFIG_GROUP_NAME}__{AllBatchConfigKeys.JOB_NAME}", "team_a_job_name"),
            (f"AIRFLOW__{team_name}___{CONFIG_GROUP_NAME}__{AllBatchConfigKeys.REGION_NAME}", "us-west-2"),
        ]
        with patch("os.environ", {key.upper(): value for key, value in config_overrides}):
            team_executor = AwsBatchExecutor(team_name=team_name)
            submit_kwargs = batch_executor_config.build_submit_kwargs(team_executor.conf)

            assert submit_kwargs["jobQueue"] == "team_a_job_queue"
            assert submit_kwargs["jobDefinition"] == "team_a_job_def"
            assert submit_kwargs["jobName"] == "team_a_job_name"

            # Now create an executor without a team and ensure the non-team configs are used.
            non_team_executor = AwsBatchExecutor()
            submit_kwargs = batch_executor_config.build_submit_kwargs(non_team_executor.conf)

            assert submit_kwargs["jobQueue"] == "some-job-queue"
            assert submit_kwargs["jobDefinition"] == "some-job-def"
            assert submit_kwargs["jobName"] == "some-job-name"


class TestBatchExecutorConfig:
    @staticmethod
    def _unset_conf():
        for env in os.environ:
            if env.startswith(f"AIRFLOW__{CONFIG_GROUP_NAME.upper()}__"):
                os.environ.pop(env)

    def teardown_method(self) -> None:
        self._unset_conf()

    @pytest.mark.skipif(
        RUNNING_TESTS_AGAINST_AIRFLOW_PACKAGES,
        reason="Config defaults are validated against provider.yaml so this test "
        "should only run when tests are run from sources",
    )
    def test_validate_config_defaults(self):
        """Assert that the defaults stated in the config.yml file match those in utils.CONFIG_DEFAULTS."""
        from airflow.providers.amazon import __file__ as provider_path

        config_filename = os.path.join(os.path.dirname(provider_path), "provider.yaml")

        with open(config_filename) as config:
            options = yaml.safe_load(config)["config"][CONFIG_GROUP_NAME]["options"]
            file_defaults = {
                option: default for (option, value) in options.items() if (default := value.get("default"))
            }

        assert len(file_defaults) == len(CONFIG_DEFAULTS)
        for key in file_defaults.keys():
            assert file_defaults[key] == CONFIG_DEFAULTS[key]

    @pytest.mark.parametrize(
        "bad_config",
        [
            pytest.param({"command": "bad_robot"}, id="executor_config_can_not_overwrite_command"),
        ],
    )
    def test_executor_config_exceptions(self, bad_config, mock_executor):
        with pytest.raises(ValueError, match='Executor Config should never override "command'):
            mock_executor.execute_async(mock_airflow_key, mock_cmd, executor_config=bad_config)

    def test_config_defaults_are_applied(self):
        submit_kwargs = batch_executor_config.build_submit_kwargs(conf)
        found_keys = {convert_camel_to_snake(key): key for key in submit_kwargs.keys()}

        for expected_key, expected_value in CONFIG_DEFAULTS.items():
            # these config options are used by the executor,
            # but are not expected to appear in the submit_kwargs.
            if expected_key in [
                AllBatchConfigKeys.AWS_CONN_ID,
                AllBatchConfigKeys.CHECK_HEALTH_ON_STARTUP,
                AllBatchConfigKeys.MAX_SUBMIT_JOB_ATTEMPTS,
                AllBatchConfigKeys.SUBMIT_JOB_KWARGS,
                AllBatchConfigKeys.REGION_NAME,
            ]:
                assert expected_key not in found_keys.keys()
            else:
                assert expected_key in found_keys.keys()
                assert expected_value == submit_kwargs[found_keys[expected_key]]

    def test_verify_tags_are_used_as_provided(self):
        """Confirm that the ``tags`` provided are not converted to camelCase."""
        templated_tags = {"Apache": "Airflow"}

        provided_run_submit_kwargs = {
            "tags": templated_tags,  # The user should be allowed to pass arbitrary submit_job args
        }

        run_submit_kwargs_env_key = (
            f"AIRFLOW__{CONFIG_GROUP_NAME}__{AllBatchConfigKeys.SUBMIT_JOB_KWARGS}".upper()
        )
        os.environ[run_submit_kwargs_env_key] = json.dumps(provided_run_submit_kwargs)
        submit_kwargs = batch_executor_config.build_submit_kwargs(conf)

        # Verify that tag names are exempt from the camel-case conversion.
        assert submit_kwargs["tags"] == templated_tags

    @pytest.mark.parametrize(
        ("submit_job_kwargs", "exec_config", "expected_result"),
        [
            # No input submit_job_kwargs or executor overrides
            (
                {},
                {},
                {
                    "jobDefinition": "some-job-def",
                    "jobQueue": "some-job-queue",
                    "jobName": "some-job-name",
                    "containerOverrides": {
                        "command": ["command"],
                        "environment": [{"name": "AIRFLOW_IS_EXECUTOR_CONTAINER", "value": "true"}],
                    },
                },
            ),
            # submit_job_kwargs provided, not exec_config
            (
                {
                    "shareIdentifier": "Banana",
                    "tags": [{"key": "FOO", "value": "BAR"}],
                    "containerOverrides": {
                        "memory": 500,
                        "vcpus": 10,
                        "environment": [{"name": "X", "value": "Y"}],
                    },
                },
                {},
                {
                    "shareIdentifier": "Banana",
                    "tags": [{"key": "FOO", "value": "BAR"}],
                    "jobDefinition": "some-job-def",
                    "jobQueue": "some-job-queue",
                    "jobName": "some-job-name",
                    "containerOverrides": {
                        "command": ["command"],
                        "memory": 500,
                        "vcpus": 10,
                        "environment": [
                            {"name": "X", "value": "Y"},
                            # Added by the batch executor
                            {"name": "AIRFLOW_IS_EXECUTOR_CONTAINER", "value": "true"},
                        ],
                    },
                },
            ),
            # exec_config provided, no submit_job_kwargs
            (
                {},
                {
                    "shareIdentifier": "Banana",
                    "tags": [{"key": "FOO", "value": "BAR"}],
                    "containerOverrides": {
                        "memory": 500,
                        "vcpus": 10,
                        "environment": [{"name": "X", "value": "Y"}],
                    },
                },
                {
                    "shareIdentifier": "Banana",
                    "tags": [{"key": "FOO", "value": "BAR"}],
                    "jobDefinition": "some-job-def",
                    "jobQueue": "some-job-queue",
                    "jobName": "some-job-name",
                    "containerOverrides": {
                        "command": ["command"],
                        "memory": 500,
                        "vcpus": 10,
                        "environment": [
                            {"name": "X", "value": "Y"},
                            # Added by the batch executor
                            {"name": "AIRFLOW_IS_EXECUTOR_CONTAINER", "value": "true"},
                        ],
                    },
                },
            ),
            # Both submit_job_kwargs and executor_config provided. The latter should override the former,
            # following a recursive python dict update strategy
            (
                {
                    "shareIdentifier": "Banana",
                    "tags": [{"key": "FOO", "value": "BAR"}],
                    "propagateTags": True,
                    "containerOverrides": {
                        "memory": 500,
                        "vcpus": 10,
                        "environment": [{"name": "X", "value": "Y"}],
                    },
                },
                {
                    "shareIdentifier": "Fish",
                    "tags": [{"key": "X", "value": "Y"}, {"key": "W", "value": "Z"}],
                    "containerOverrides": {
                        "memory": 300,
                        "environment": [{"name": "W", "value": "Z"}],
                    },
                },
                {
                    # tags and shareIdentifier are overridden by exec_config
                    "shareIdentifier": "Fish",
                    # List types overwrite entirely, as python dict update would do
                    "tags": [{"key": "X", "value": "Y"}, {"key": "W", "value": "Z"}],
                    # propagateTags remains since it is not a list type and not overridden by exec config
                    "propagateTags": True,
                    "jobDefinition": "some-job-def",
                    "jobQueue": "some-job-queue",
                    "jobName": "some-job-name",
                    "containerOverrides": {
                        "command": ["command"],
                        "memory": 300,
                        # vcpus is present because it is missing from the exec config
                        "vcpus": 10,
                        "environment": [
                            # Overridden list type
                            {"name": "W", "value": "Z"},  # Only new env vars present, overwritten
                            # Added by the batch executor
                            {"name": "AIRFLOW_IS_EXECUTOR_CONTAINER", "value": "true"},
                        ],
                    },
                },
            ),
        ],
    )
    def test_submit_job_kwargs_exec_config_overrides(
        self, set_env_vars, submit_job_kwargs, exec_config, expected_result
    ):
        submit_job_kwargs_env_key = (
            f"AIRFLOW__{CONFIG_GROUP_NAME}__{AllBatchConfigKeys.SUBMIT_JOB_KWARGS}".upper()
        )
        os.environ[submit_job_kwargs_env_key] = json.dumps(submit_job_kwargs)

        mock_ti_key = mock.Mock(spec=tuple)
        command = ["command"]

        executor = AwsBatchExecutor()

        final_run_task_kwargs = executor._submit_job_kwargs(mock_ti_key, command, "queue", exec_config)

        assert final_run_task_kwargs == expected_result

    def test_short_import_path(self):
        from airflow.providers.amazon.aws.executors.batch import AwsBatchExecutor as AwsBatchExecutorShortPath

        assert AwsBatchExecutor is AwsBatchExecutorShortPath



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
from collections.abc import Callable
from datetime import datetime, timedelta

try:
    from airflow.sdk import timezone
except ImportError:
    from airflow.utils import timezone  # type: ignore[attr-defined,no-redef]

log = logging.getLogger(__name__)


def calculate_next_attempt_delay(
    attempt_number: int,
    max_delay: int = 60 * 2,
    exponent_base: int = 4,
) -> timedelta:
    """
    Calculate the exponential backoff (in seconds) until the next attempt.

    :param attempt_number: Number of attempts since last success.
    :param max_delay: Maximum delay in seconds between retries. Default 120.
    :param exponent_base: Exponent base to calculate delay. Default 4.
    """
    return timedelta(seconds=min((exponent_base**attempt_number), max_delay))


def exponential_backoff_retry(
    last_attempt_time: datetime,
    attempts_since_last_successful: int,
    callable_function: Callable,
    max_delay: int = 60 * 2,
    max_attempts: int = -1,
    exponent_base: int = 4,
) -> None:
    """
    Retry a callable function with exponential backoff between attempts if it raises an exception.

    :param last_attempt_time: Timestamp of last attempt call.
    :param attempts_since_last_successful: Number of attempts since last success.
    :param callable_function: Callable function that will be called if enough time has passed.
    :param max_delay: Maximum delay in seconds between retries. Default 120.
    :param max_attempts: Maximum number of attempts before giving up. Default -1 (no limit).
    :param exponent_base: Exponent base to calculate delay. Default 4.
    """
    if max_attempts != -1 and attempts_since_last_successful >= max_attempts:
        log.error("Max attempts reached. Exiting.")
        return

    next_retry_time = last_attempt_time + calculate_next_attempt_delay(
        attempt_number=attempts_since_last_successful, max_delay=max_delay, exponent_base=exponent_base
    )

    current_time = timezone.utcnow()

    if current_time >= next_retry_time:
        try:
            callable_function()
        except Exception:
            log.exception("Error calling %r", callable_function.__name__)
            next_delay = calculate_next_attempt_delay(
                attempts_since_last_successful + 1, max_delay, exponent_base
            )
            log.info("Waiting for %s seconds before retrying.", next_delay)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import boto3
import pytest
from moto import mock_aws

from airflow.models import DAG
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.transfers.s3_to_sftp import S3ToSFTPOperator
from airflow.providers.ssh.hooks.ssh import SSHHook
from airflow.providers.ssh.operators.ssh import SSHOperator
from airflow.utils.timezone import datetime

from tests_common.test_utils.config import conf_vars

pytestmark = pytest.mark.db_test


TASK_ID = "test_s3_to_sftp"
BUCKET = "test-s3-bucket"
S3_KEY = "test/test_1_file.csv"
SFTP_PATH = "/tmp/remote_path.txt"
SFTP_CONN_ID = "ssh_default"
LOCAL_FILE_PATH = "/tmp/test_s3_upload"

SFTP_MOCK_FILE = "test_sftp_file.csv"
S3_MOCK_FILES = "test_1_file.csv"

TEST_DAG_ID = "unit_tests_s3_to_sftp"
DEFAULT_DATE = datetime(2018, 1, 1)


class TestS3ToSFTPOperator:
    def setup_method(self):
        hook = SSHHook(ssh_conn_id="ssh_default")
        hook.no_host_key_check = True
        dag = DAG(
            f"{TEST_DAG_ID}test_schedule_dag_once",
            start_date=DEFAULT_DATE,
            schedule="@once",
        )

        self.hook = hook

        self.ssh_client = self.hook.get_conn()
        self.sftp_client = self.ssh_client.open_sftp()

        self.dag = dag
        self.s3_bucket = BUCKET
        self.sftp_path = SFTP_PATH
        self.s3_key = S3_KEY

    @mock_aws
    @conf_vars({("core", "enable_xcom_pickling"): "True"})
    def test_s3_to_sftp_operation(self):
        s3_hook = S3Hook(aws_conn_id=None)
        # Setting
        test_remote_file_content = (
            "This is remote file content \n which is also multiline "
            "another line here \n this is last line. EOF"
        )

        # Test for creation of s3 bucket
        conn = boto3.client("s3")
        conn.create_bucket(Bucket=self.s3_bucket)
        assert s3_hook.check_for_bucket(self.s3_bucket)

        with open(LOCAL_FILE_PATH, "w") as file:
            file.write(test_remote_file_content)
        s3_hook.load_file(LOCAL_FILE_PATH, self.s3_key, bucket_name=BUCKET)

        # Check if object was created in s3
        objects_in_dest_bucket = conn.list_objects(Bucket=self.s3_bucket, Prefix=self.s3_key)
        # there should be object found, and there should only be one object found
        assert len(objects_in_dest_bucket["Contents"]) == 1

        # the object found should be consistent with dest_key specified earlier
        assert objects_in_dest_bucket["Contents"][0]["Key"] == self.s3_key

        # get remote file to local
        run_task = S3ToSFTPOperator(
            s3_bucket=BUCKET,
            s3_key=S3_KEY,
            sftp_path=SFTP_PATH,
            sftp_conn_id=SFTP_CONN_ID,
            task_id=TASK_ID,
            dag=self.dag,
        )
        assert run_task is not None

        run_task.execute(None)

        # Check that the file is created remotely
        check_file_task = SSHOperator(
            task_id="test_check_file",
            ssh_hook=self.hook,
            command=f"cat {self.sftp_path}",
            do_xcom_push=True,
            dag=self.dag,
        )
        assert check_file_task is not None
        result = check_file_task.execute(None)
        assert result.strip() == test_remote_file_content.encode("utf-8")

        # Clean up after finishing with test
        conn.delete_object(Bucket=self.s3_bucket, Key=self.s3_key)
        conn.delete_bucket(Bucket=self.s3_bucket)
        assert not s3_hook.check_for_bucket(self.s3_bucket)

    def delete_remote_resource(self):
        # Initiate SHH hook
        hook = SSHHook(ssh_conn_id="ssh_default")
        hook.no_host_key_check = True
        # check the remote file content
        remove_file_task = SSHOperator(
            task_id="test_rm_file",
            ssh_hook=hook,
            command=f"rm {self.sftp_path}",
            do_xcom_push=True,
            dag=self.dag,
        )
        assert remove_file_task is not None
        remove_file_task.execute(None)

    @mock_aws
    @conf_vars({("core", "enable_xcom_pickling"): "True"})
    def test_s3_to_sftp_operation_confirm_true_default(self):
        """Test that S3ToSFTPOperator works with confirm=True by default (real SSH connection)"""
        s3_hook = S3Hook(aws_conn_id=None)
        # Setting
        test_remote_file_content = (
            "This is remote file content for confirm=True test \n which is also multiline "
            "another line here \n this is last line. EOF"
        )

        # Test for creation of s3 bucket
        conn = boto3.client("s3")
        conn.create_bucket(Bucket=self.s3_bucket)
        assert s3_hook.check_for_bucket(self.s3_bucket)

        with open(LOCAL_FILE_PATH, "w") as file:
            file.write(test_remote_file_content)
        s3_hook.load_file(LOCAL_FILE_PATH, self.s3_key, bucket_name=BUCKET)

        # Check if object was created in s3
        objects_in_dest_bucket = conn.list_objects(Bucket=self.s3_bucket, Prefix=self.s3_key)
        # there should be object found, and there should only be one object found
        assert len(objects_in_dest_bucket["Contents"]) == 1

        # the object found should be consistent with dest_key specified earlier
        assert objects_in_dest_bucket["Contents"][0]["Key"] == self.s3_key

        # get remote file to local - Test with default confirm=True
        run_task = S3ToSFTPOperator(
            s3_bucket=BUCKET,
            s3_key=S3_KEY,
            sftp_path=SFTP_PATH,
            sftp_conn_id=SFTP_CONN_ID,
            task_id=TASK_ID + "_confirm_true",
            dag=self.dag,
        )
        assert run_task is not None

        run_task.execute(None)

        # Check that the file is created remotely with correct content
        check_file_task = SSHOperator(
            task_id="test_check_file_confirm_true",
            ssh_hook=self.hook,
            command=f"cat {self.sftp_path}",
            do_xcom_push=True,
            dag=self.dag,
        )
        assert check_file_task is not None
        result = check_file_task.execute(None)
        assert result.strip() == test_remote_file_content.encode("utf-8")

        # Clean up after finishing with test
        conn.delete_object(Bucket=self.s3_bucket, Key=self.s3_key)
        conn.delete_bucket(Bucket=self.s3_bucket)
        assert not s3_hook.check_for_bucket(self.s3_bucket)

    @mock_aws
    @conf_vars({("core", "enable_xcom_pickling"): "True"})
    def test_s3_to_sftp_operation_confirm_false(self):
        """Test that S3ToSFTPOperator works with confirm=False when specified (real SSH connection)"""
        s3_hook = S3Hook(aws_conn_id=None)
        # Setting
        test_remote_file_content = (
            "This is remote file content for confirm=False test \n which is also multiline "
            "another line here \n this is last line. EOF"
        )

        # Test for creation of s3 bucket
        conn = boto3.client("s3")
        conn.create_bucket(Bucket=self.s3_bucket)
        assert s3_hook.check_for_bucket(self.s3_bucket)

        with open(LOCAL_FILE_PATH, "w") as file:
            file.write(test_remote_file_content)
        s3_hook.load_file(LOCAL_FILE_PATH, self.s3_key, bucket_name=BUCKET)

        # Check if object was created in s3
        objects_in_dest_bucket = conn.list_objects(Bucket=self.s3_bucket, Prefix=self.s3_key)
        # there should be object found, and there should only be one object found
        assert len(objects_in_dest_bucket["Contents"]) == 1

        # the object found should be consistent with dest_key specified earlier
        assert objects_in_dest_bucket["Contents"][0]["Key"] == self.s3_key

        # get remote file to local - Test with explicit confirm=False
        run_task = S3ToSFTPOperator(
            s3_bucket=BUCKET,
            s3_key=S3_KEY,
            sftp_path=SFTP_PATH,
            sftp_conn_id=SFTP_CONN_ID,
            task_id=TASK_ID + "_confirm_false",
            confirm=False,  # Explicitly set to False
            dag=self.dag,
        )
        assert run_task is not None

        run_task.execute(None)

        # Check that the file is created remotely with correct content
        check_file_task = SSHOperator(
            task_id="test_check_file_confirm_false",
            ssh_hook=self.hook,
            command=f"cat {self.sftp_path}",
            do_xcom_push=True,
            dag=self.dag,
        )
        assert check_file_task is not None
        result = check_file_task.execute(None)
        assert result.strip() == test_remote_file_content.encode("utf-8")

        # Clean up after finishing with test
        conn.delete_object(Bucket=self.s3_bucket, Key=self.s3_key)
        conn.delete_bucket(Bucket=self.s3_bucket)
        assert not s3_hook.check_for_bucket(self.s3_bucket)

    def teardown_method(self):
        self.delete_remote_resource()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest
from sqlalchemy import delete, or_

from airflow import models
from airflow.providers.amazon.aws.transfers.s3_to_sql import S3ToSqlOperator
from airflow.providers.common.compat.sdk import AirflowException
from airflow.utils.session import create_session

pytestmark = pytest.mark.db_test


class TestS3ToSqlTransfer:
    @pytest.fixture(autouse=True)
    def setup_connections(self, create_connection_without_db):
        create_connection_without_db(
            models.Connection(
                conn_id="s3_test",
                conn_type="aws",
                schema="test",
                extra='{"aws_access_key_id": "aws_access_key_id", "aws_secret_access_key":'
                ' "aws_secret_access_key"}',
            )
        )
        create_connection_without_db(
            models.Connection(
                conn_id="sql_test",
                conn_type="postgres",
                host="some.host.com",
                schema="test_db",
                login="user",
                password="password",
            )
        )

        self.s3_to_sql_transfer_kwargs = {
            "task_id": "s3_to_sql_task",
            "aws_conn_id": "s3_test",
            "sql_conn_id": "sql_test",
            "s3_key": "test/test.csv",
            "s3_bucket": "testbucket",
            "table": "sql_table",
            "column_list": ["Column1", "Column2"],
            "schema": "sql_schema",
            "commit_every": 5000,
        }

    @pytest.fixture
    def mock_parser(self):
        return MagicMock()

    @pytest.fixture
    def mock_bad_hook(self):
        bad_hook = MagicMock()
        del bad_hook.insert_rows
        return bad_hook

    @patch("airflow.providers.amazon.aws.transfers.s3_to_sql.NamedTemporaryFile")
    @patch("airflow.models.connection.Connection.get_hook")
    @patch("airflow.providers.amazon.aws.transfers.s3_to_sql.S3Hook.get_key")
    def test_execute(self, mock_get_key, mock_hook, mock_tempfile, mock_parser):
        S3ToSqlOperator(parser=mock_parser, **self.s3_to_sql_transfer_kwargs).execute({})

        mock_get_key.assert_called_once_with(
            key=self.s3_to_sql_transfer_kwargs["s3_key"],
            bucket_name=self.s3_to_sql_transfer_kwargs["s3_bucket"],
        )

        mock_get_key.return_value.download_fileobj.assert_called_once_with(
            mock_tempfile.return_value.__enter__.return_value
        )

        mock_parser.assert_called_once_with(mock_tempfile.return_value.__enter__.return_value.name)

        mock_hook.return_value.insert_rows.assert_called_once_with(
            table=self.s3_to_sql_transfer_kwargs["table"],
            schema=self.s3_to_sql_transfer_kwargs["schema"],
            target_fields=self.s3_to_sql_transfer_kwargs["column_list"],
            rows=mock_parser.return_value,
            commit_every=self.s3_to_sql_transfer_kwargs["commit_every"],
        )

    @patch("airflow.providers.amazon.aws.transfers.s3_to_sql.NamedTemporaryFile")
    @patch("airflow.models.connection.Connection.get_hook", return_value=mock_bad_hook)
    @patch("airflow.providers.amazon.aws.transfers.s3_to_sql.S3Hook.get_key")
    def test_execute_with_bad_hook(self, mock_get_key, mock_bad_hook, mock_tempfile, mock_parser):
        with pytest.raises(AirflowException):
            S3ToSqlOperator(parser=mock_parser, **self.s3_to_sql_transfer_kwargs).execute({})

    def test_hook_params(self, mock_parser):
        op = S3ToSqlOperator(
            parser=mock_parser,
            sql_hook_params={
                "log_sql": False,
            },
            **self.s3_to_sql_transfer_kwargs,
        )
        hook = op.db_hook
        assert hook.log_sql == op.sql_hook_params["log_sql"]

    def teardown_method(self):
        with create_session() as session:
            session.execute(
                delete(models.Connection).where(
                    or_(models.Connection.conn_id == "s3_test", models.Connection.conn_id == "sql_test")
                )
            )
